"""Module for camera aligners."""

from dataclasses import dataclass

import numpy as np


@dataclass(frozen=True)
class RigidTransformation:
    """Class representing a rigid transformation (SO(3), SE(3), Sim(3)."""

    scale: float
    rotation: np.ndarray
    translation: np.ndarray

    def apply(self, points: np.ndarray) -> np.ndarray:
        """
        Applies the rigid transformation to the points.
        :arg points: shape NxM - M = point dim, N = point count
        """
        # transformed: np.ndarray = self.scale * self.points @ self.rotation + self.translation
        transformed: np.ndarray = (
            self.scale * self.rotation.dot(points.T).T + self.translation
        )
        return transformed


def align_points_arun(A: np.ndarray, B: np.ndarray) -> RigidTransformation:
    """
    Estimates SE(3) transformation between two point sets using Arun's method.
    Estimates R and t such that B ~ RA + t

    Matching points should correspond with array indices, such that:
    A[:, i] ~ B[:, i]

    :arg A: source point set - shape (3, N)
    :arg B: target point set - shape (3, N)
    """
    assert A.shape == B.shape, "point sets A and B must be of equal size"

    N: int = A.shape[1]

    # Calculate centroids as (M, 1) vectors
    A_centroid: np.ndarray = np.mean(A, axis=0)
    B_centroid: np.ndarray = np.mean(B, axis=0)

    # Calculate the points relative to the centroid
    A_prime: np.ndarray = A - A_centroid
    B_prime: np.ndarray = B - B_centroid

    # Estimate rotation
    H: np.ndarray = np.zeros([3, 3])
    for i in range(N):
        ai: np.ndarray = A_prime[i, :]
        bi: np.ndarray = B_prime[i, :]
        H: np.ndarray = H + np.outer(ai, bi)

    U: np.ndarray
    S: np.ndarray
    V_transpose: np.ndarray
    U, S, V_transpose = np.linalg.svd(H)

    V: np.ndarray = np.transpose(V_transpose)
    U_transpose: np.ndarray = np.transpose(U)
    R: np.ndarray = (
        V @ np.diag([1, 1, np.linalg.det(V) * np.linalg.det(U_transpose)]) @ U_transpose
    )

    # Estimate translation by rotating around the centroid
    t: np.ndarray = B_centroid - R @ A_centroid

    return RigidTransformation(scale=1.0, rotation=R, translation=t)


def align_points_umeyama(
    source_points: np.ndarray, target_points: np.ndarray
) -> RigidTransformation:
    """
    Estimates the Sim(3) transformation between two point sets using Umeyamas method.
    Estimates c, R and t such as B ~ c * R @ A.T + t.

    :arg A: source point set - shape (N, M)
    :arg B: target point set - shape (N, M)

    :return c: float
    :return R: numpy.array - shape (M, M)
    :return t: numpy.array - shape (M, 1)
    """
    assert (
        source_points.shape == target_points.shape
    ), "source and target point sets must be of equal size"

    A: np.ndarray = source_points
    B: np.ndarray = target_points

    point_count: int = A.shape[0]
    point_dim: int = A.shape[1]

    # Calculate the mean point of the two point sets
    mu_a: np.ndarray = A.mean(axis=0)
    mu_b: np.ndarray = B.mean(axis=0)

    # Shape var_a = ()
    var_a: np.ndarray = np.square(A - mu_a).sum(axis=1).mean()

    # Shape cov_ab = (3, 3)
    cov_ab: np.ndarray = ((B - mu_b).T @ (A - mu_a)) / point_count

    # Expected shapes:   U = (3, 3), D = (3, 1), VT = (3, 3)
    U, D, VT = np.linalg.svd(cov_ab)
    S = np.eye(point_dim)

    if np.linalg.det(U) * np.linalg.det(VT) < 0:
        S[-1, -1] = -1

    scale: float = np.trace(np.diag(D) @ S) / var_a
    rotation: np.ndarray = U @ S @ VT
    translation: np.ndarray = mu_b - scale * rotation @ mu_a

    return RigidTransformation(scale, rotation, translation)



"""Module for image data."""

import typing

from collections.abc import Callable
from dataclasses import dataclass
from enum import StrEnum, auto

from typing import Optional, Self

import numpy as np


class PixelFormat(StrEnum):
    """Class representing an image format."""

    GRAY = auto()
    X = auto()
    RGB = auto()
    BGR = auto()
    XYZ = auto()
    RGBA = auto()
    BGRA = auto()


@dataclass(frozen=True)
class ImageLayout:
    """Class representing an image layout."""

    height: int
    width: int
    channels: int

    @property
    def shape(self: Self) -> tuple[int, int, int]:
        """Returns the shape of the image layout."""
        return (self.height, self.width, self.channels)


# Create a generic variable that can be 'Parent', or any subclass.
T: typing.TypeVar = typing.TypeVar("T", bound="Image")


@dataclass(frozen=True)
class Image:
    """Class representing an image."""

    Format: typing.ClassVar[typing.TypeAlias] = PixelFormat
    Layout: typing.ClassVar[typing.TypeAlias] = ImageLayout

    _data: np.ndarray
    _pixel_format: PixelFormat
    _layout: ImageLayout

    @property
    def pixel_format(self: Self) -> PixelFormat:
        """Returns the pixel format of the image."""
        return self._pixel_format

    @property
    def layout(self: Self) -> ImageLayout:
        """Return the layout of the image."""
        return self._layout

    @property
    def height(self: Self) -> int:
        """Returns the height of the image."""
        return self._layout.height

    @property
    def width(self: Self) -> int:
        """Returns the height of the image."""
        return self._layout.width

    @property
    def channels(self: Self) -> int:
        """Returns the height of the image."""
        return self._layout.channels

    @property
    def shape(self: Self) -> tuple[int, int, int]:
        """Returns the shape of the image."""
        return (self.height, self.width, self.channels)

    @property
    def dtype(self: Self) -> np.dtype:
        """Returns the data type of the image."""
        return self._data.dtype

    @property
    def ndim(self: Self) -> int:
        """Returns the number of dimension of the image."""
        return self._data.ndim

    @classmethod
    def from_array(cls: type[T], data: np.ndarray, pixel_format: PixelFormat) -> T:
        """Creates an image from an array."""

        if data.ndim == 2:
            data: np.ndarray = np.expand_dims(data, axis=-1)
        elif data.ndim == 3:
            pass
        else:
            raise ValueError(f"invalid image data dimension: {data.ndim}")

        layout: ImageLayout = ImageLayout(
            height=data.shape[0], width=data.shape[1], channels=data.shape[2]
        )
        return cls(data, pixel_format, layout)

    def to_array(self: Self) -> np.ndarray:
        """Returns the image pixels as an array."""
        return self._data.copy()

    def copy(self: Self) -> T:
        """Returns a copy of the image."""
        return Image(self._data, self._pixel_format, self._layout)


ImageLoader = Callable[[None], Image]



"""Package for mynds tasks."""

from .camera_assimilation import assimilate_camera_references
from .camera_export import (
    export_cameras,
    export_stereo_cameras,
    tabularize_cameras,
    tabularize_stereo_cameras,
)

__all__ = []



"""Module with generic generator functions."""

from collections.abc import Iterable
from typing import TypeVar


T: TypeVar = TypeVar("T")


def generate_chunks(items: Iterable[T], max_size: int) -> Iterable[Iterable[T]]:
    """Generate chunks of the items with a given maximum chunk size."""
    for index in range(0, len(items), max_size):
        yield items[index : index + max_size]


def generate_chunked_items(items: Iterable[T], max_size: int) -> Iterable[Iterable[T]]:
    """Generates an iterable of item chunks with a maximum size."""
    return [items[index : index + max_size] for index in range(0, len(items), max_size)]



"""Module for testing schema models, i.e. camera, sensor, calibration, reference."""

import pytest

import numpy as np
import sqlmodel as sqlm

import mynd.schemas as schemas


"""
Calibration unit tests:
 - test_calibration_initialization
 - test_projection_matrix
 - test_distortion_vector
 - test_projection_matrix_type
 - test_distortion_vector_type
 - test_calibration_with_different_values
"""


@pytest.fixture
def sample_calibration():
    return schemas.CalibrationSchema(
        id=1,
        width=1920,
        height=1080,
        fx=1000.0,
        fy=1000.0,
        cx=960.0,
        cy=540.0,
        k1=0.1,
        k2=0.2,
        k3=0.3,
        p1=0.01,
        p2=0.02,
    )


def test_calibration_initialization(sample_calibration):
    assert isinstance(sample_calibration, sqlm.SQLModel)
    assert sample_calibration.id == 1
    assert sample_calibration.width == 1920
    assert sample_calibration.height == 1080
    assert sample_calibration.fx == 1000.0
    assert sample_calibration.fy == 1000.0
    assert sample_calibration.cx == 960.0
    assert sample_calibration.cy == 540.0
    assert sample_calibration.k1 == 0.1
    assert sample_calibration.k2 == 0.2
    assert sample_calibration.k3 == 0.3
    assert sample_calibration.p1 == 0.01
    assert sample_calibration.p2 == 0.02


def test_projection_matrix(sample_calibration):
    expected_matrix = np.array(
        [[1000.0, 0.0, 960.0], [0.0, 1000.0, 540.0], [0.0, 0.0, 1.0]]
    )
    assert np.allclose(sample_calibration.projection_matrix, expected_matrix)


def test_distortion_vector(sample_calibration):
    expected_vector = np.array([0.1, 0.2, 0.01, 0.02, 0.3])
    assert np.allclose(sample_calibration.distortion_vector, expected_vector)


def test_projection_matrix_type(sample_calibration):
    assert isinstance(sample_calibration.projection_matrix, np.ndarray)
    assert sample_calibration.projection_matrix.shape == (3, 3)


def test_distortion_vector_type(sample_calibration):
    assert isinstance(sample_calibration.distortion_vector, np.ndarray)
    assert sample_calibration.distortion_vector.shape == (5,)


def test_calibration_with_different_values():
    cal = schemas.CalibrationSchema(
        id=2,
        width=640,
        height=480,
        fx=500.0,
        fy=500.0,
        cx=320.0,
        cy=240.0,
        k1=-0.1,
        k2=0.05,
        k3=0.0,
        p1=-0.001,
        p2=0.001,
    )

    expected_matrix = np.array(
        [[500.0, 0.0, 320.0], [0.0, 500.0, 240.0], [0.0, 0.0, 1.0]]
    )
    assert np.allclose(cal.projection_matrix, expected_matrix)

    expected_vector = np.array([-0.1, 0.05, -0.001, 0.001, 0.0])
    assert np.allclose(cal.distortion_vector, expected_vector)


"""
schemas.SensorSchema unit tests:
 - test_sensor_initialization
 - test_sensor_default_values
 - test_sensor_location_vector
 - test_sensor_rotation_matrix
 - test_sensor_location_vector_type
 - test_sensor_rotation_matrix_type
 - test_sensor_with_none_location_rotation
 - test_sensor_with_different_values
"""


@pytest.fixture
def sample_sensor(sample_calibration):
    return schemas.SensorSchema(
        id=1,
        label="Test schemas.SensorSchema",
        width=1920,
        height=1080,
        location=[1.0, 2.0, 3.0],
        rotation=[[1.0, 0.0, 0.0], [0.0, 1.0, 0.0], [0.0, 0.0, 1.0]],
        calibration=sample_calibration,
    )


def test_sensor_initialization(sample_sensor, sample_calibration):
    assert isinstance(sample_sensor, sqlm.SQLModel)
    assert sample_sensor.id == 1
    assert sample_sensor.label == "Test schemas.SensorSchema"
    assert sample_sensor.width == 1920
    assert sample_sensor.height == 1080
    assert sample_sensor.location == [1.0, 2.0, 3.0]
    assert sample_sensor.rotation == [[1.0, 0.0, 0.0], [0.0, 1.0, 0.0], [0.0, 0.0, 1.0]]
    assert sample_sensor.calibration == sample_calibration


def test_sensor_default_values():
    sensor = schemas.SensorSchema(
        id=2, label="Default schemas.SensorSchema", width=640, height=480
    )
    assert sensor.location is None
    assert sensor.rotation is None
    assert sensor.calibration is None


def test_sensor_location_vector(sample_sensor):
    expected_vector = np.array([1.0, 2.0, 3.0])
    assert np.allclose(sample_sensor.location_vector, expected_vector)


def test_sensor_rotation_matrix(sample_sensor):
    expected_matrix = np.array([[1.0, 0.0, 0.0], [0.0, 1.0, 0.0], [0.0, 0.0, 1.0]])
    assert np.allclose(sample_sensor.rotation_matrix, expected_matrix)


def test_sensor_location_vector_type(sample_sensor):
    assert isinstance(sample_sensor.location_vector, np.ndarray)
    assert sample_sensor.location_vector.shape == (3,)


def test_sensor_rotation_matrix_type(sample_sensor):
    assert isinstance(sample_sensor.rotation_matrix, np.ndarray)
    assert sample_sensor.rotation_matrix.shape == (3, 3)


def test_sensor_with_none_location_rotation():
    sensor = schemas.SensorSchema(id=3, label="None Sensor", width=800, height=600)
    assert sensor.location_vector is None
    assert sensor.rotation_matrix is None


def test_sensor_with_different_values():
    sensor = schemas.SensorSchema(
        id=4,
        label="Different Sensor",
        width=3840,
        height=2160,
        location=[5.0, -2.0, 1.5],
        rotation=[[0.0, -1.0, 0.0], [1.0, 0.0, 0.0], [0.0, 0.0, 1.0]],
    )
    assert np.allclose(sensor.location_vector, np.array([5.0, -2.0, 1.5]))
    assert np.allclose(
        sensor.rotation_matrix,
        np.array([[0.0, -1.0, 0.0], [1.0, 0.0, 0.0], [0.0, 0.0, 1.0]]),
    )


"""
Reference unit tests:
 - test_reference_initialization
 - test_reference_location_vector
 - test_reference_rotation_vector
 - test_reference_location_vector_type
 - test_reference_rotation_vector_type
 - test_reference_with_different_values
 - test_reference_property_immutability
 - test_reference_attributes
 - test_reference_str_representation
"""


@pytest.fixture
def sample_reference():
    return schemas.ReferenceSchema(
        id=1,
        epsg_code=4326,
        longitude=10.123,
        latitude=50.456,
        height=100.5,
        yaw=30.0,
        pitch=5.0,
        roll=0.5,
    )


def test_reference_initialization(sample_reference):
    assert isinstance(sample_reference, sqlm.SQLModel)
    assert sample_reference.id == 1
    assert sample_reference.epsg_code == 4326
    assert sample_reference.longitude == 10.123
    assert sample_reference.latitude == 50.456
    assert sample_reference.height == 100.5
    assert sample_reference.yaw == 30.0
    assert sample_reference.pitch == 5.0
    assert sample_reference.roll == 0.5


def test_reference_location_vector(sample_reference):
    expected_vector = np.array([10.123, 50.456, 100.5])
    assert np.allclose(sample_reference.location_vector, expected_vector)


def test_reference_rotation_vector(sample_reference):
    expected_vector = np.array([30.0, 5.0, 0.5])
    assert np.allclose(sample_reference.rotation_vector, expected_vector)


def test_reference_location_vector_type(sample_reference):
    assert isinstance(sample_reference.location_vector, np.ndarray)
    assert sample_reference.location_vector.shape == (3,)


def test_reference_rotation_vector_type(sample_reference):
    assert isinstance(sample_reference.rotation_vector, np.ndarray)
    assert sample_reference.rotation_vector.shape == (3,)


def test_reference_with_different_values():
    ref = schemas.ReferenceSchema(
        id=2,
        epsg_code=3857,
        longitude=-74.006,
        latitude=40.7128,
        height=10.0,
        yaw=180.0,
        pitch=-10.0,
        roll=2.5,
    )
    assert np.allclose(ref.location_vector, np.array([-74.006, 40.7128, 10.0]))
    assert np.allclose(ref.rotation_vector, np.array([180.0, -10.0, 2.5]))


def test_reference_property_immutability(sample_reference):
    location_vector = sample_reference.location_vector
    rotation_vector = sample_reference.rotation_vector

    # Attempt to modify the vectors
    location_vector[0] = 0
    rotation_vector[0] = 0

    # Check that the original properties are unchanged
    assert np.allclose(
        sample_reference.location_vector, np.array([10.123, 50.456, 100.5])
    )
    assert np.allclose(sample_reference.rotation_vector, np.array([30.0, 5.0, 0.5]))


@pytest.mark.parametrize(
    "attr",
    [
        "id",
        "epsg_code",
        "longitude",
        "latitude",
        "height",
        "yaw",
        "pitch",
        "roll",
    ],
)
def test_reference_attributes(sample_reference, attr):
    assert hasattr(sample_reference, attr)


def test_reference_str_representation(sample_reference):
    assert str(sample_reference) != ""
    assert repr(sample_reference) != ""



"""Module for mynds schema models."""

from collections.abc import Callable
from typing import ClassVar, Optional, TypeAlias, Self

import numpy as np

from sqlmodel import SQLModel, Field


class ChunkGroupSchema(SQLModel):
    """Class representing a chunk group schema."""

    id: int | None = Field(default=None)
    label: str

    chunks: list["ChunkSchema"] = Field(default_factory=list)


class ChunkSchema(SQLModel):
    """Class representing a chunk schema."""

    id: int | None = Field(default=None)
    label: str
    meta: dict = Field(default_factory=dict)

    sensors: list["SensorSchema"] = Field(default_factory=list)
    cameras: list["CameraSchema"] = Field(default_factory=list)

    stereo_rigs: list["StereoRigSchema"] = Field(default_factory=list)
    stereo_sensor_pairs: list["StereoSensorPairSchema"] = Field(default_factory=list)
    stereo_camera_pairs: list["StereoCameraPairSchema"] = Field(default_factory=list)


class CalibrationSchema(SQLModel):
    """Class representing a calibration schema."""

    id: int | None = Field(default=None)
    width: int
    height: int
    fx: float
    fy: float
    cx: float
    cy: float
    k1: float
    k2: float
    k3: float
    p1: float
    p2: float

    @property
    def focal_length(self) -> float:
        """Returns the focal length of the calibration."""
        return self.fx

    @property
    def projection_matrix(self) -> np.ndarray:
        """Returns the calibration projection matrix."""
        return np.array(
            [[self.fx, 0.0, self.cx], [0.0, self.fy, self.cy], [0.0, 0.0, 1.0]]
        )

    @property
    def distortion_vector(self) -> np.ndarray:
        """Returns the calibration distortion vector."""
        return np.array([self.k1, self.k2, self.p1, self.p2, self.k3])


class SensorBaseSchema(SQLModel):
    """Class representing a sensor base schema."""

    id: int | None = Field(default=None)
    label: str
    width: int
    height: int

    location: list[float] | None = Field(default=None)
    rotation: list[list[float]] | None = Field(default=None)

    calibration: CalibrationSchema | None = Field(default=None)

    def __hash__(self) -> hash:
        """Returns a hash for the object."""
        return hash(self.id)

    @property
    def location_vector(self) -> np.ndarray | None:
        """Returns the sensor location as a numpy array."""
        if not self.location:
            return None
        return np.array(self.location)

    @property
    def rotation_matrix(self) -> np.ndarray | None:
        """Returns the sensor rotation as a numpy array."""
        if not self.rotation:
            return None
        return np.array(self.rotation)


class RectifiedSensorSchema(SensorBaseSchema):
    """Class representing a rectified sensor schema."""

    pass


class SensorSchema(SensorBaseSchema):
    """Class representing a sensor schema."""

    rectified_sensor: Optional[RectifiedSensorSchema] = Field(default=None)

    @property
    def rectified(self) -> RectifiedSensorSchema | None:
        """Gets the rectified sensor."""
        return self.rectified_sensor

    @rectified.setter
    def rectified(self, sensor: RectifiedSensorSchema) -> None:
        """Sets the rectified sensor."""
        self.rectified_sensor = sensor

    def is_rectified(self) -> bool:
        """Returns true if the sensor is rectified."""
        return self.rectified_sensor is not None


class ReferenceSchema(SQLModel):
    """Class representing a reference schema."""

    id: int | None = Field(default=None)
    epsg_code: int

    longitude: float
    latitude: float
    height: float

    yaw: float
    pitch: float
    roll: float

    @property
    def location_vector(self) -> np.ndarray:
        """Returns the reference location as a numpy array."""
        return np.array([self.longitude, self.latitude, self.height])

    @property
    def rotation_vector(self) -> np.ndarray:
        """Return the reference rotation as a numpy array."""
        return np.array([self.yaw, self.pitch, self.roll])


class CameraSchema(SQLModel):
    """Class representing a camera schema."""

    SensorType: ClassVar[TypeAlias] = SensorSchema
    ReferenceType: ClassVar[TypeAlias] = ReferenceSchema

    id: int | None = Field(default=None)
    label: str
    image_label: str
    meta: dict = Field(default_factory=dict)

    sensor: SensorSchema
    prior_reference: ReferenceSchema | None = Field(default=None)
    aligned_reference: ReferenceSchema | None = Field(default=None)
    assimilated_reference: ReferenceSchema | None = Field(default=None)


class PixelMapSchema(SQLModel):
    """Class representing a pixel map schema. The pixel maps is represented as
    floating point lists of shape HxWx2."""

    id: int | None = Field(default=None)
    values: list[list[list[float]]]

    @property
    def shape(self) -> tuple[int, ...]:
        """Returns the shape of the pixel map."""
        return self.to_array().shape

    @property
    def height(self) -> int:
        """Returns the height of the pixel map."""
        return self.shape[0]

    @property
    def width(self) -> int:
        """Returns the width of the pixel map."""
        return self.shape[1]

    @property
    def x(self) -> np.ndarray:
        """Returns the x component of the pixel map."""
        return self.to_array()[:, :, 0]

    @property
    def y(self) -> np.ndarray:
        """Returns the y component of the pixel map."""
        return self.to_array()[:, :, 1]

    def to_array(self) -> np.ndarray:
        """Returns the pixel map as an array."""
        return np.array(self.values).astype(np.float32)


class StereoSensorComponentSchema(SQLModel):
    """Class representing a stereo sensor component."""

    id: int | None = Field(default=None)
    sensor: SensorSchema

    @property
    def rectified_sensor(self) -> RectifiedSensorSchema | None:
        """Gets the rectified sensor record."""
        return self.sensor.rectified_sensor

    @rectified_sensor.setter
    def rectified_sensor(self, sensor: RectifiedSensorSchema) -> None:
        """Sets the rectified sensor record."""
        self.sensor.rectified_sensor = sensor

    def is_rectified(self) -> bool:
        """Returns true if the stereo sensor component is rectified."""
        return self.sensor.rectified is not None


class StereoSensorPairSchema(SQLModel):
    """Class representing a stereo sensor pair schema."""

    Component: ClassVar[TypeAlias] = StereoSensorComponentSchema

    id: int | None = Field(default=None)
    master_component: Component
    slave_component: Component

    @property
    def master(self) -> Component:
        """Gets the master component."""
        return self.master_component

    @master.setter
    def master(self, component: Component) -> None:
        """Sets the master component."""
        self.master_component = component

    @property
    def slave(self) -> Component:
        """Gets the slave component."""
        return self.slave_component

    @slave.setter
    def slave(self, component: Component) -> None:
        """Sets the slave component."""
        self.slave_component = component

    @property
    def baseline(self) -> float:
        """Returns the baseline between the master and slave sensor."""
        relative_location: np.ndarray = (
            self.slave.sensor.location_vector - self.master.sensor.location_vector
        )
        return np.linalg.norm(relative_location)


class StereoCameraPairSchema(SQLModel):
    """Class representing a stereo camera pair schema."""

    id: int | None = Field(default=None)
    master: CameraSchema
    slave: CameraSchema
    stereo_rig: Optional["StereoRigSchema"] = Field(default=None)


class StereoPixelMapSchema(SQLModel):
    """Class representing a pixel map pair schema."""

    id: int | None = Field(default=None)
    master: PixelMapSchema
    slave: PixelMapSchema


class StereoRigBase(SQLModel):
    """Class representing a stereo rig base."""

    # Define class vars with type aliases for ease of access
    SensorComponent: ClassVar[TypeAlias] = StereoSensorComponentSchema
    SensorPair: ClassVar[TypeAlias] = StereoSensorPairSchema
    CameraPair: ClassVar[TypeAlias] = StereoCameraPairSchema
    PixelMapPair: ClassVar[TypeAlias] = StereoPixelMapSchema

    id: int | None = Field(default=None)
    sensors: StereoSensorPairSchema


class StereoRigSchema(StereoRigBase):
    """Class representing a stereo rig schema without pixel maps."""

    pass


class StereoRigWithMapsSchema(StereoRigBase):
    """Class representing a stereo rig schema with pixel maps"""

    pixel_maps: StereoPixelMapSchema | None = Field(default=None)



"""Module for registration pipeline functionality."""

from collections.abc import Callable
from dataclasses import dataclass, field
from typing import TypeAlias

from mynd.geometry.point_cloud import PointCloud, PointCloudProcessor

from .types import (
    RigidTransformation,
    RegistrationResult,
    PointCloudAligner,
    PointCloudRefiner,
)


@dataclass
class RegistrationPipeline:
    """Class representing a registration pipeline."""

    @dataclass
    class AlignerModule:
        """Class representing an initializer module."""

        preprocessor: PointCloudProcessor
        registrator: PointCloudAligner

    @dataclass
    class RefinerModule:
        """Class representing an incrementor module."""

        preprocessor: PointCloudProcessor
        registrator: PointCloudRefiner

    initializer: AlignerModule
    incrementors: list[RefinerModule] = field(default_factory=list)

    Callback = TypeAlias = Callable[[PointCloud, PointCloud, RegistrationResult], None]


Pipeline: TypeAlias = RegistrationPipeline


def apply_registration_pipeline(
    pipeline: Pipeline,
    source: PointCloud,
    target: PointCloud,
    callback: Pipeline.Callback | None = None,
) -> RegistrationResult:
    """Applies a registration pipeline to the source and target."""

    result: RegistrationResult = _apply_aligner_module(
        pipeline.initializer,
        source=source,
        target=target,
        callback=callback,
    )

    for incrementor in pipeline.incrementors:
        result: RegistrationResult = _apply_refiner_module(
            incrementor,
            source=source,
            target=target,
            transformation=result.transformation,
            callback=callback,
        )

    return result


def _apply_aligner_module(
    module: Pipeline.AlignerModule,
    source: PointCloud,
    target: PointCloud,
    callback: Pipeline.Callback | None = None,
) -> RegistrationResult:
    """Applies an initializer module to the source and target."""

    source_pre: PointCloud = module.preprocessor(source)
    target_pre: PointCloud = module.preprocessor(target)

    result: RegistrationResult = module.registrator(
        source=source_pre, target=target_pre
    )

    if callback is not None:
        callback(source_pre, target_pre, result)

    return result


def _apply_refiner_module(
    module: Pipeline.RefinerModule,
    source: PointCloud,
    target: PointCloud,
    transformation: RigidTransformation,
    callback: Pipeline.Callback | None = None,
) -> RegistrationResult:
    """Applies an incrementor module to the source and target."""

    source_pre: PointCloud = module.preprocessor(source)
    target_pre: PointCloud = module.preprocessor(target)

    result: RegistrationResult = module.registrator(
        source=source_pre,
        target=target_pre,
        transformation=transformation,
    )

    if callback is not None:
        callback(source_pre, target_pre, result)

    return result



"""Package for image type and processors."""

from .color_conversion import (
    convert_image_color,
    convert_image_to_rgb,
    apply_color_map,
)

from .image_io import (
    read_image,
    write_image,
)

from .image_processors import (
    flip_image,
    resize_image,
    process_image_clahe,
    process_image_gamma,
    process_image_linear,
    normalize_image,
)

from .image_types import (
    Image,
    ImageLayout,
    PixelFormat,
    ImageLoader,
)

__all__ = [
    "convert_image_color",
    "convert_image_to_rgb",
    "apply_color_map",
    # ...
    "read_image",
    "write_image",
    # ...
    "flip_image",
    "resize_image",
    "process_image_clahe",
    "process_image_gamma",
    "process_image_linear",
    "normalize_image",
    # ...
    "Image",
    "ImageLayout",
    "PixelFormat",
    "ImageLoader",
]



"""Module for logging functionality."""

import sys
import pathlib

import loguru


LOG_LEVEL: str = "DEBUG"

LOG_FORMAT = (
    "<green>{time:YYYY-MM-DD HH:mm:ss.SSS}</green>"
    " | <level>{level: <4}</level>"
    " | <cyan>Line {line: >4} ({file}):</cyan> <b>{message}</b>"
)


def add_file_sink(file: pathlib.Path) -> None:
    """Adds a file sink to the logger."""
    loguru.logger.add(
        file,
        level=LOG_LEVEL,
        format=LOG_FORMAT,
        colorize=False,
        backtrace=True,
        diagnose=True,
    )


def _initialize() -> None:
    """Initializes the logger."""

    # Clear default logger
    loguru.logger.remove()

    # Add custom sinks
    loguru.logger.add(
        sys.stderr,
        level=LOG_LEVEL,
        format=LOG_FORMAT,
        colorize=True,
        backtrace=True,
        diagnose=True,
    )

    return loguru.logger


logger = _initialize()
global_logger = logger



"""Package with ORM record models."""

import sqlalchemy as sqla
import sqlmodel as sqlm

import mynd.database as db

from .records import ChunkGroupRecord as ChunkGroupRecord
from .records import ChunkRecord as ChunkRecord

from .records import CalibrationRecord as CalibrationRecord
from .records import RectifiedCalibrationRecord as RectifiedCalibrationRecord

from .records import SensorRecord as SensorRecord
from .records import RectifiedSensorRecord as RectifiedSensorRecord

from .records import PriorReferenceRecord as PriorReferenceRecord
from .records import AlignedReferenceRecord as AlignedReferenceRecord
from .records import AssimilatedReferenceRecord as AssimilatedReferenceRecord

from .records import CameraRecord as CameraRecord
from .records import PixelMapRecord as PixelMapRecord

from .records import StereoSensorComponentRecord as StereoSensorComponentRecord
from .records import StereoSensorRecord as StereoSensorRecord
from .records import StereoCameraPairRecord as StereoCameraPairRecord
from .records import StereoRigRecord as StereoRigRecord


def close_all_sessions() -> None:
    """Closes all SQLAlchemy sessions."""
    sqla.orm.session.close_all_sessions()


def create_database_tables(engine: db.Engine) -> None:
    """Creates database tables based on the schema of the imported SQL models."""
    sqlm.SQLModel.metadata.create_all(engine)


def clear_database_tables(engine: db.Engine) -> None:
    """Clears the SQLModel tables in a database."""
    close_all_sessions()
    sqlm.SQLModel.metadata.drop_all(engine)


def get_database_tables() -> dict:
    """Returns the registered SQL models."""
    tables: dict = {
        name: table for name, table in sqlm.SQLModel.metadata.tables.items()
    }
    return tables


__all__ = []



"""Module for registration data types."""

from collections.abc import Callable
from dataclasses import dataclass
from typing import TypeAlias

import numpy as np
import open3d

from mynd.geometry.point_cloud import PointCloud


@dataclass
class RegistrationResult:
    """Class representing registration results including the information matrix."""

    fitness: float
    inlier_rmse: float
    correspondence_set: np.ndarray
    transformation: np.ndarray
    information: np.ndarray


# Feature registration interfaces
Feature: TypeAlias = open3d.pipelines.registration.Feature
FeatureExtractor: TypeAlias = Callable[[PointCloud], Feature]
FeatureMatcher: TypeAlias = Callable[
    [PointCloud, PointCloud, Feature, Feature], RegistrationResult
]

# Point cloud registration interfaces
RigidTransformation: TypeAlias = np.ndarray
PointCloudAligner: TypeAlias = Callable[[PointCloud, PointCloud], RegistrationResult]
PointCloudRefiner: TypeAlias = Callable[
    [PointCloud, PointCloud, RigidTransformation], RegistrationResult
]



"""Module with generic custom containers."""

from collections.abc import Iterator
from dataclasses import dataclass, field
from typing import TypeVar, Generic, Optional


T = TypeVar("T")


@dataclass(frozen=True)
class Pair(Generic[T]):
    """Class representing a pair of items of the same type."""

    first: T
    second: T

    def __iter__(self) -> Iterator:
        """Returns an iterator for the pair."""
        return iter((self.first, self.second))


Key = TypeVar("Key")
Value = TypeVar("Value")


@dataclass
class Registry(Generic[Key, Value]):
    """Generic registry class."""

    _items: dict[Key, Value] = field(default_factory=dict)

    def __len__(self) -> int:
        """Returns the number of items in the registry."""
        return len(self._items)

    def __contains__(self, key: Key) -> bool:
        """Returns the number of items in the registry."""
        return key in self._items

    def __getitem__(self, key: Key) -> Value:
        """Returns the value for the given key."""
        return self._items[key]

    def __setitem__(self, key: Key, value: Value) -> None:
        """Inserts a key value pair in the registry."""
        self._items[key] = value

    @property
    def count(self) -> int:
        """Returns the number of items in the registry."""
        return len(self)

    @property
    def values(self) -> list[Value]:
        """Returns the values for the registered items."""
        return list(self._items.values())

    @property
    def keys(self) -> list[Key]:
        """Returns the keys for the registered items."""
        return list(self._items.keys())

    def get(self, key: Key, default: Optional[Value] = None) -> Value:
        """Returns a value if it is in the registry, or the default value."""
        return self._items.get(key, default)

    def items(self) -> list[tuple[Key, Value]]:
        """Returns the items in the registry."""
        return self._items.items()

    def insert(self, key: Key, value: Value) -> None:
        """Insert a key value pair in the registry."""
        self._items[key] = value

    def remove(self, key: Key) -> None:
        """Removes a key value pair in the registry."""
        self._items.pop(key)

    def pop(self, key: Key) -> Value:
        """Insert a key value pair in the registry."""
        return self._items.pop(key)



"""Module for stereo visualization functionality."""

import copy
import functools
import time

from abc import abstractmethod
from concurrent.futures import ThreadPoolExecutor
from dataclasses import dataclass
from typing import Optional

import numpy as np
import cv2

import open3d as o3d
import open3d.visualization.gui as gui
import open3d.visualization.rendering as rendering


# TODO: Implement these objects
IntParameter = object
EnumParameter = object
StereoOutput = object
StereoMethod = object
Calibration = object
InputPair = object


disparity_window = None


class ImageWindow:
    def __init__(self, name: str, size: tuple[int, int]):
        self.name = name
        self.window = gui.Application.instance.create_window(name, size[0], size[1])
        self.image_widget = gui.ImageWidget()
        self.window.add_child(self.image_widget)

    def update_image(self, image: np.ndarray):
        image_geom = o3d.geometry.Image(image)
        self.image_widget.update_image(image_geom)
        self.window.post_redraw()


class ImageWindowsManager:
    def __init__(self):
        self.windows_by_name = {}

    def render_image(
        self,
        name: str,
        image: np.ndarray,
        window_title: Optional[str],
        max_size: int,
    ):
        if name not in self.windows_by_name:
            rows, cols, _ = image.shape
            if cols > rows:
                initial_size = max_size, int(max_size * rows / cols)
            else:
                initial_size = int(max_size * cols / rows), max_size
            self.windows_by_name[name] = ImageWindow(name, initial_size)
        self.windows_by_name[name].update_image(image)
        if window_title is not None:
            self.windows_by_name[name].title = window_title


# TODO:
@functools.lru_cache
def get_image_window_manager() -> ImageWindowsManager:
    """Returns the image manager. When first called, the function initializes
    the image manager and caches the instance."""
    return ImageWindowsManager()


def render_image(
    name: str, image: np.ndarray, window_title: str | None = None, max_size: int = 640
) -> None:
    image_windows_manager: ImageWindowsManager = get_image_window_manager()
    assert isinstance(
        image_windows_manager, ImageWindowsManager
    ), "image window manager not initialized"
    image_windows_manager.render_image(name, image, window_title, max_size)


def color_disparity(disparity_map: np.ndarray, calibration: Calibration):
    min_disp = (calibration.fx * calibration.baseline_meters) / calibration.depth_range[
        1
    ]
    # disparity_pixels = (calibration.fx * calibration.baseline_meters) / depth_meters
    max_disp = (calibration.fx * calibration.baseline_meters) / calibration.depth_range[
        0
    ]
    norm_disparity_map = 255 * ((disparity_map - min_disp) / (max_disp - min_disp))
    disparity_color = cv2.applyColorMap(
        cv2.convertScaleAbs(norm_disparity_map, 1), cv2.COLORMAP_VIRIDIS
    )
    return disparity_color


def show_color_disparity(name: str, color_disparity: np.ndarray):
    render_image("StereoDemo - Disparity", color_disparity, name)


class Settings:
    def __init__(self):
        self.show_axes = False


class Source:
    def __init__(self):
        pass

    @abstractmethod
    def is_live(self) -> bool:
        """Whether the source is capture live images or not"""
        return False

    def selected_index(self) -> int:
        return 0

    @abstractmethod
    def get_next_pair(self) -> InputPair:
        return InputPair(None, None, None, None)

    def get_pair_at_index(self, idx: int) -> InputPair:
        return InputPair(None, None, None, None)

    def get_pair_list(self) -> list[str]:
        return []


class StereoVisualizer:
    def __init__(self, stereo_methods: dict[str, StereoMethod], source: Source):
        gui.Application.instance.initialize()

        self.vis = gui.Application.instance
        self.source = source

        self.executor = ThreadPoolExecutor(max_workers=1)
        self.executor_future = None
        self._progress_dialog = None
        self._last_progress_update_time = None

        self.stereo_methods = stereo_methods
        self.stereo_methods_output = {}
        self.input = InputPair(None, None, None, None)
        self._downsample_factor = 0

        self.window = gui.Application.instance.create_window("StereoVis", 1280, 1024)
        w = self.window  # to make the code more concise

        self.settings = Settings()

        # 3D widget
        self._scene = gui.SceneWidget()
        self._scene.scene = rendering.Open3DScene(w.renderer)
        # self._scene.scene.show_ground_plane(True, rendering.Scene.GroundPlane.XZ)
        self._scene.set_view_controls(gui.SceneWidget.Controls.ROTATE_CAMERA)
        self._scene.set_on_key(self._on_key_pressed)

        self._clear_outputs()

        for name, o in self.stereo_methods_output.items():
            if o.point_cloud is not None:
                self._scene.scene.add_geometry(
                    name, o.point_cloud, rendering.MaterialRecord()
                )

        self._reset_camera()

        em = w.theme.font_size
        self.separation_height = int(round(0.5 * em))
        self._settings_panel = gui.Vert(
            0, gui.Margins(0.25 * em, 0.25 * em, 0.25 * em, 0.25 * em)
        )

        self._next_image_button = gui.Button("Next Image")
        self._next_image_button.set_on_clicked(self._next_image_clicked)
        self._settings_panel.add_child(self._next_image_button)

        if not self.source.is_live():
            self._settings_panel.add_fixed(self.separation_height)
            self.images_combo = gui.Combobox()
            input_pairs = self.source.get_pair_list()
            for pair_name in input_pairs:
                self.images_combo.add_item(pair_name)
            self.images_combo.selected_index = 0
            self.images_combo.set_on_selection_changed(self._image_selected)
            self._settings_panel.add_child(self.images_combo)
            self._settings_panel.add_fixed(self.separation_height)
        else:
            self.images_combo = None

        horiz = gui.Horiz(0, gui.Margins(0.25 * em, 0.25 * em, 0.25 * em, 0.25 * em))
        label = gui.Label("Input downsampling")
        label.tooltip = "Number of /2 downsampling steps to apply on the input"
        horiz.add_child(label)
        downsampling_slider = gui.Slider(gui.Slider.INT)
        downsampling_slider.set_limits(0, 4)
        downsampling_slider.int_value = self._downsample_factor
        downsampling_slider.set_on_value_changed(self._downsampling_changed)
        horiz.add_child(downsampling_slider)
        self._settings_panel.add_child(horiz)

        self._settings_panel.add_fixed(self.separation_height)

        self.algo_list = gui.ListView()
        self.algo_list.set_items(list(stereo_methods.keys()))
        self.algo_list.selected_index = 0
        self.algo_list.set_max_visible_items(8)
        self.algo_list.set_on_selection_changed(self._on_algo_list_selected)
        self._settings_panel.add_child(self.algo_list)

        self.method_params_proxy = gui.WidgetProxy()
        self._settings_panel.add_child(self.method_params_proxy)

        self.last_runtime = gui.Label("")
        self._settings_panel.add_child(self.last_runtime)

        self.input_status = gui.Label("No input.")
        self._settings_panel.add_child(self.input_status)

        view_ctrls = gui.CollapsableVert(
            "View controls", 0.25 * em, gui.Margins(em, 0, 0, 0)
        )
        reset_cam_button = gui.Button("Reset Camera")
        reset_cam_button.set_on_clicked(self._reset_camera)
        view_ctrls.add_child(reset_cam_button)

        horiz = gui.Horiz(0, gui.Margins(0.25 * em, 0.25 * em, 0.25 * em, 0.25 * em))
        label = gui.Label("Max depth (m)")
        label.tooltip = "Max depth to render in meters"
        horiz.add_child(label)
        self.depth_range_slider = gui.Slider(gui.Slider.DOUBLE)
        self.depth_range_slider.set_limits(0.5, 1000)
        self.depth_range_slider.double_value = 100
        self.depth_range_slider.set_on_value_changed(self._depth_range_slider_changed)
        horiz.add_child(self.depth_range_slider)
        view_ctrls.add_child(horiz)

        self._depth_range_manually_changed = False

        self._settings_panel.add_fixed(self.separation_height)
        self._settings_panel.add_child(view_ctrls)

        w.set_on_layout(self._on_layout)
        w.add_child(self._scene)
        w.add_child(self._settings_panel)

        self._on_algo_list_selected(self.algo_list.selected_value, False)
        self._apply_settings()

        if self.source.is_live():
            self.read_next_pair()
        else:
            self._image_selected(None, None)

    def _on_key_pressed(self, keyEvent) -> gui.SceneWidget.EventCallbackResult:
        if keyEvent.key == gui.KeyName.Q:
            self.vis.quit()
            return gui.SceneWidget.EventCallbackResult.HANDLED
        return gui.SceneWidget.EventCallbackResult.IGNORED

    def _downsampling_changed(self, v):
        self._downsample_factor = int(v)
        self._process_input(self.full_res_input)

    def _downsample_input(self, input: InputPair):
        for i in range(0, self._downsample_factor):
            if np.max(input.left_image.shape[:2]) < 250:
                break
            input.left_image = cv2.pyrDown(input.left_image)
            input.right_image = cv2.pyrDown(input.right_image)
            if input.input_disparity is not None:
                input.input_disparity = cv2.pyrDown(input.input_disparity)
            input.calibration.downsample(
                input.left_image.shape[1], input.left_image.shape[0]
            )

    def read_next_pair(self):
        input = self.source.get_next_pair()
        self._update_pair_index()
        self._process_input(input)

    def _process_input(self, input):
        if self._downsample_factor > 0:
            self.full_res_input = input
            input = copy.deepcopy(input)
            self._downsample_input(input)
        else:
            self.full_res_input = input

        if not self._depth_range_manually_changed:
            self.depth_range_slider.double_value = input.calibration.depth_range[1]

        render_image(
            "StereoDemo - Input image",
            np.hstack([input.left_image, input.right_image]),
        )

        self.input = input
        self.input_status.text = (
            f"Input: {input.left_image.shape[1]}x{input.left_image.shape[0]} "
            + input.status
        )

        if self.input.has_data():
            assert (
                self.input.left_image.shape[1] == self.input.calibration.width
                and self.input.left_image.shape[0] == self.input.calibration.height
            )
            self.o3dCameraIntrinsic = o3d.camera.PinholeCameraIntrinsic(
                width=self.input.left_image.shape[1],
                height=self.input.left_image.shape[0],
                fx=self.input.calibration.fx,
                fy=self.input.calibration.fy,
                cx=self.input.calibration.cx0,
                cy=self.input.calibration.cy,
            )

            self._clear_outputs()
            self._run_current_method()

    def update_once(self):
        if self.executor_future is not None:
            self._check_run_complete()
        return gui.Application.instance.run_one_tick()

    def _clear_outputs(self):
        for name in self.stereo_methods.keys():
            self.stereo_methods_output[name] = StereoOutput(
                disparity_pixels=None,
                color_image_bgr=None,
                computation_time=np.nan,
            )
            if self._scene.scene.has_geometry(name):
                self._scene.scene.remove_geometry(name)

    def _reset_camera(self):
        # bbox = o3d.geometry.AxisAlignedBoundingBox(np.array([-10, 0,-10]), np.array([0,3,0]))
        bbox = self._scene.scene.bounding_box
        min_bound, max_bound = bbox.min_bound.copy(), bbox.max_bound.copy()
        min_bound[0] = min(min_bound[0], -5)
        min_bound[2] = min(min_bound[2], -5)
        max_bound[0] = max(max_bound[0], 5)
        max_bound[1] = max(max_bound[1], 2)
        max_bound[2] = 0
        bbox.min_bound, bbox.max_bound = min_bound, max_bound

        self._scene.setup_camera(60.0, bbox, np.array([0, 0, 0]))
        eye = np.array([0, 0.5, 1.0])
        lookat = np.array([0, 0, -1.0])
        up = np.array([0, 1.0, 0])
        self._scene.look_at(lookat, eye, up)

        if self.input.has_data():
            self._depth_range_manually_changed = False
            self.depth_range_slider.double_value = self.input.calibration.depth_range[1]
            self._update_rendering()

    def _build_stereo_method_widgets(self, name):
        em = self.window.theme.font_size
        method = self.stereo_methods[name]
        container = gui.Vert(0, gui.Margins(0.25 * em, 0.25 * em, 0.25 * em, 0.25 * em))
        label = gui.Label(method.description)
        label.text_color = gui.Color(1.0, 0.5, 0.0)
        container.add_child(label)
        self._reload_settings_functions = []
        for name, param in method.parameters.items():
            if isinstance(param, IntParameter):
                horiz = gui.Horiz(
                    0, gui.Margins(0.25 * em, 0.25 * em, 0.25 * em, 0.25 * em)
                )
                label = gui.Label(name)
                label.tooltip = param.description
                horiz.add_child(label)
                slider = gui.Slider(gui.Slider.INT)
                slider.set_limits(param.min, param.max)
                slider.int_value = param.value

                def set_value_from_method(slider=slider, method=method, name=name):
                    slider.int_value = method.parameters[name].value

                self._reload_settings_functions.append(set_value_from_method)

                # workaround late binding
                def callback(value, method=method, name=name, slider=slider):
                    p = method.parameters[name]
                    p.set_value(int(value))
                    slider.int_value = p.value

                slider.set_on_value_changed(callback)
                horiz.add_child(slider)
                container.add_child(horiz)
            elif isinstance(param, EnumParameter):
                horiz = gui.Horiz(
                    0, gui.Margins(0.25 * em, 0.25 * em, 0.25 * em, 0.25 * em)
                )
                label = gui.Label(name)
                label.tooltip = param.description
                horiz.add_child(label)
                combo = gui.Combobox()
                for value in param.values:
                    combo.add_item(value)
                combo.selected_index = param.index

                def callback(
                    combo_idx, combo_val, method=method, name=name, combo=combo
                ):
                    method.parameters[name].set_index(combo.selected_index)

                combo.set_on_selection_changed(callback)

                def set_value_from_method(combo=combo, method=method, name=name):
                    combo.selected_index = method.parameters[name].index

                self._reload_settings_functions.append(set_value_from_method)
                horiz.add_child(combo)
                container.add_child(horiz)

        horiz = gui.Horiz(0, gui.Margins(0.25 * em, 0.25 * em, 0.25 * em, 0.25 * em))
        apply_button = gui.Button("Apply")
        apply_button.horizontal_padding_em = 3
        apply_button.set_on_clicked(self._run_current_method)
        horiz.add_child(apply_button)
        horiz.add_fixed(self.separation_height)
        reset_default = gui.Button("Reset defaults")
        reset_default.set_on_clicked(self._reset_method_defaults)
        horiz.add_child(reset_default)
        container.add_child(horiz)
        return container

    def _on_algo_list_selected(self, name: str, is_dbl_click: bool):
        self.method_params_proxy.set_widget(self._build_stereo_method_widgets(name))
        self._update_runtime()
        for other_name in self.stereo_methods_output.keys():
            self._scene.scene.show_geometry(other_name, False)
        self._scene.scene.show_geometry(name, True)
        self._apply_settings()
        if self.stereo_methods_output[name].disparity_pixels is None:
            self._run_current_method()
        if self.stereo_methods_output[name].disparity_color is not None:
            show_color_disparity(name, self.stereo_methods_output[name].disparity_color)

    def _on_show_axes(self, show):
        self.settings.show_axes = show
        self._apply_settings()

    def _next_image_clicked(self):
        self.read_next_pair()

    def _image_selected(self, combo_idx, combo_val):
        idx = self.images_combo.selected_index
        input = self.source.get_pair_at_index(idx)
        self._process_input(input)

    def _update_pair_index(self):
        if self.images_combo is not None:
            self.images_combo.selected_index = self.source.selected_index()

    def _apply_settings(self):
        self._scene.scene.show_axes(self.settings.show_axes)

    def _reset_method_defaults(self):
        name = self.algo_list.selected_value
        method = self.stereo_methods[name]
        method.reset_defaults()
        for callable in self._reload_settings_functions:
            callable()

    def _check_run_complete(self) -> None:
        if not self.executor_future.done():
            if self._progress_dialog is None:
                self._progress_dialog: ProgressDialogData = create_progress_dialog(
                    "Running the current method",
                    f"Computing {self.algo_list.selected_value}...",
                    self.window.theme.font_size,
                )

                self.window.show_dialog(self._progress_dialog.dialog)

            now = time.time()
            if now - self._last_progress_update_time > 0.1:
                self._last_progress_update_time = now
                self._run_progress.value += (1.0 - self._run_progress.value) / 16.0
            return

        if self._progress_dialog:
            self.window.close_dialog()
        self._progress_dialog = None

        stereo_output = self.executor_future.result()
        self.executor_future = None

        x0, y0, x1, y1 = self.input.calibration.left_image_rect_normalized
        x0 = int(x0 * stereo_output.disparity_pixels.shape[1] + 0.5)
        x1 = int(x1 * stereo_output.disparity_pixels.shape[1] + 0.5)
        y0 = int(y0 * stereo_output.disparity_pixels.shape[0] + 0.5)
        y1 = int(y1 * stereo_output.disparity_pixels.shape[0] + 0.5)
        valid_mask = np.zeros(stereo_output.disparity_pixels.shape, dtype=np.uint8)
        valid_mask[y0:y1, x0:x1] = 1
        stereo_output.disparity_pixels[valid_mask == 0] = -1.0

        name = self.algo_list.selected_value
        stereo_output.disparity_color = color_disparity(
            stereo_output.disparity_pixels, self.input.calibration
        )
        show_color_disparity(name, stereo_output.disparity_color)

        self.stereo_methods_output[name] = stereo_output
        self._update_rendering([name])
        self._update_runtime()

    def _depth_range_slider_changed(self, v: float):
        self._depth_range_manually_changed = True
        self._update_rendering()

    def _update_rendering(self, names_to_update=None):
        if names_to_update is None:
            names_to_update = list(self.stereo_methods_output.keys())

        selected_name = self.algo_list.selected_value

        for name in names_to_update:
            stereo_output = self.stereo_methods_output[name]
            if stereo_output.disparity_pixels is None:
                continue

            depth_meters = StereoMethod.depth_meters_from_disparity(
                stereo_output.disparity_pixels, self.input.calibration
            )

            if self._scene.scene.has_geometry(name):
                self._scene.scene.remove_geometry(name)

            o3d_color = o3d.geometry.Image(
                cv2.cvtColor(stereo_output.color_image_bgr, cv2.COLOR_BGR2RGB)
            )
            o3d_depth = o3d.geometry.Image(depth_meters)
            rgbd = o3d.geometry.RGBDImage.create_from_color_and_depth(
                o3d_color,
                o3d_depth,
                1,
                depth_trunc=self.depth_range_slider.int_value,
                convert_rgb_to_intensity=False,
            )
            stereo_output.point_cloud = o3d.geometry.PointCloud.create_from_rgbd_image(
                rgbd, self.o3dCameraIntrinsic
            )
            stereo_output.point_cloud.transform(
                [[1, 0, 0, 0], [0, -1, 0, 0], [0, 0, -1, 0], [0, 0, 0, 1]]
            )
            self._scene.scene.add_geometry(
                name, stereo_output.point_cloud, rendering.MaterialRecord()
            )
            self._scene.scene.show_geometry(name, name == selected_name)

    def _run_current_method(self):
        if self.executor_future is not None:
            return self._check_run_complete()

        if not self.input.has_data():
            return

        name = self.algo_list.selected_value

        def do_beefy_work():
            stereo_output = self.stereo_methods[name].compute_disparity(self.input)
            return stereo_output

        self._last_progress_update_time = time.time()
        self.executor_future = self.executor.submit(do_beefy_work)

    def _update_runtime(self) -> None:
        name = self.algo_list.selected_value
        output = self.stereo_methods_output[name]
        if np.isnan(output.computation_time):
            self.last_runtime.text = "No output yet."
        else:
            self.last_runtime.text = (
                f"Computation time: {output.computation_time*1e3:.1f} ms"
            )

    def _on_layout(self, layout_context) -> None:
        """Sets the frame, i.e. position and size, for every child."""
        settings_width = 17 * layout_context.theme.font_size
        rectangle = self.window.content_rect
        self._scene.frame = gui.Rect(
            0,
            rectangle.y,
            rectangle.get_right() - settings_width,
            rectangle.height,
        )
        height = rectangle.height
        self._settings_panel.frame = gui.Rect(
            rectangle.get_right() - settings_width,
            rectangle.y,
            settings_width,
            height,
        )


@dataclass
class ProgressDialogData:
    dialog: gui.Dialog
    progress_bar: gui.ProgressBar


def create_progress_dialog(title: str, message: str, margin: int) -> ProgressDialogData:
    # A Dialog is just a widget, so you make its child a layout just like
    # a Window.
    dialog = gui.Dialog(title)

    dialog_layout: gui.Vert = gui.Vert(
        margin, gui.Margins(margin, margin, margin, margin)
    )
    dialog_layout.add_child(gui.Label(message))

    # Add the Ok button. We need to define a callback function to handle
    # the click.
    progress_bar: gui.ProgressBar = gui.ProgressBar()
    progress_bar.value = 0.1  # 10% complete
    prog_layout = gui.Horiz(margin)
    prog_layout.add_child(progress_bar)
    dialog_layout.add_child(prog_layout)

    dialog.add_child(dialog_layout)

    return ProgressDialogData(dialog, progress_bar)



"""Module for geometric image transformations."""

from dataclasses import dataclass
from typing import Self

import cv2
import numpy as np

from mynd.image import Image


@dataclass(frozen=True)
class PixelMap:
    """Class representing a pixel map. The pixel map data is an
    array of size HxWx2 with the X and Y pixel maps respectively."""

    data: np.ndarray

    def __post_init__(self: Self) -> None:
        """Validates the pixel map."""
        assert self.data.dtype in [np.float32, float], "invalid pixel map dtype"
        assert self.data.ndim == 3, "invalid pixel map dimensions"
        assert self.data.shape[2] == 2, "invalid pixel map shape"

    @property
    def height(self: Self) -> int:
        """Returns the height of the pixel map."""
        return self.data.shape[0]

    @property
    def width(self: Self) -> int:
        """Returns the width of the pixel map."""
        return self.data.shape[1]

    @property
    def shape(self: Self) -> tuple[int, int, int]:
        """Returns the shape of the pixel map."""
        return self.data.shape

    @property
    def ndim(self: Self) -> int:
        """Returns the dimension count of the pixel map."""
        return self.data.ndim

    @property
    def dtype(self: Self) -> type:
        """Returns the data type of the pixel map."""
        return self.data.dtype

    @property
    def x(self: Self) -> np.ndarray:
        """Returns the x component of the pixel map."""
        return self.data[:, :, 0]

    @property
    def y(self: Self) -> np.ndarray:
        """Returns the y component of the pixel map."""
        return self.data[:, :, 1]

    def to_array(self: Self) -> np.ndarray:
        """Returns the pixel map as an array."""
        return self.data.copy()


def compute_pixel_map(
    camera_matrix: np.ndarray,
    distortion: np.ndarray,
    rotation: np.ndarray,
    new_camera_matrix: np.ndarray,
    desired_resolution: tuple[int, int],
) -> PixelMap:
    """Computes a pixel map that maps the image as seen by original camera matrix,
    to an image seen by the new camera matrix."""

    map_components: tuple[np.ndarray, np.ndarray] = cv2.initUndistortRectifyMap(
        camera_matrix,
        distortion,
        rotation,
        new_camera_matrix,
        desired_resolution,
        cv2.CV_32FC1,
    )

    return PixelMap(np.stack((map_components[0], map_components[1]), axis=-1))


def invert_pixel_map(
    pixel_map: PixelMap,
    *,
    iterations: int = 20,
    step_size: float = 0.5,
) -> PixelMap:
    """Computes the inverse of a pixel map using iteration. The function
    takes a HxWx2 array representing a map from indices to subpixel index,
    and returns a HxWx2 array representing the inverse map."""

    height: int = pixel_map.height
    width: int = pixel_map.width

    identity = np.zeros_like(pixel_map.data)
    identity[:, :, 1], identity[:, :, 0] = np.indices((height, width))  # identity map
    inverse_map_data: np.ndarray = np.copy(identity)

    for index in range(iterations):
        correction: np.ndarray = identity - cv2.remap(
            src=pixel_map.data,
            map1=inverse_map_data,
            map2=None,
            borderMode=cv2.BORDER_DEFAULT,
            interpolation=cv2.INTER_LINEAR,
        )
        inverse_map_data += correction * step_size

    return PixelMap(inverse_map_data)


def remap_image_pixels(
    image: Image,
    pixel_map: PixelMap,
    *,
    border_value: int | float = 0,
    border_mode: int = cv2.BORDER_CONSTANT,
    interpolation: int = cv2.INTER_LINEAR,
) -> Image:
    """Applies a pixel map to the pixels of the image."""
    mapped: np.ndarray = cv2.remap(
        src=image.to_array(),
        map1=pixel_map.to_array(),
        map2=None,
        borderValue=border_value,
        borderMode=border_mode,
        interpolation=interpolation,
    )

    if mapped.ndim == 2:
        mapped: np.ndarray = np.expand_dims(mapped, axis=2)

    return Image.from_array(data=mapped, pixel_format=image.pixel_format)



"""Module for point cloud processors."""

from copy import deepcopy

import open3d

from .point_cloud_types import PointCloud, PointCloudProcessor


def downsample_point_cloud(
    cloud: PointCloud,
    spacing: float,
    inplace: bool = False,
) -> PointCloud:
    """Downsamples a point cloud by performing voxel resampling."""
    if not inplace:
        cloud = deepcopy(cloud)

    cloud = cloud.voxel_down_sample(voxel_size=spacing)
    return cloud


def estimate_point_cloud_normals(
    cloud: PointCloud,
    radius: float = 0.10,
    neighbours: int = 30,
    inplace: bool = False,
) -> PointCloud:
    """Estimates the normals of a point cloud based on neighbouring points."""
    if not inplace:
        cloud = deepcopy(cloud)

    cloud.estimate_normals(
        search_param=open3d.geometry.KDTreeSearchParamHybrid(radius=0.1, max_nn=30)
    )
    return cloud


def create_downsampler(
    spacing: float,
    inplace: bool = False,
) -> PointCloudProcessor:
    def downsampler_wrapper(
        cloud: PointCloud,
    ) -> PointCloud:
        """Wraps a point cloud downsampler."""
        return downsample_point_cloud(cloud=cloud, spacing=spacing, inplace=inplace)

    return downsampler_wrapper


def create_normal_estimator(
    radius: float = 0.10,
    neighbours: int = 30,
    inplace: bool = False,
) -> PointCloudProcessor:
    def normal_estimator_wrapper(cloud: PointCloud) -> PointCloud:
        """Wraps a point cloud normal estimator."""
        return estimate_point_cloud_normals(
            cloud=cloud,
            radius=radius,
            neighbours=neighbours,
            inplace=inplace,
        )

    return normal_estimator_wrapper



"""Module with database functionality for Mynds ORM."""

import typing

import sqlalchemy as sqla
import sqlmodel as sqlm
import sqlalchemy.ext.asyncio as sqlasync

import mynd.utils.env as env


Engine: typing.TypeAlias = sqla.Engine
Session: typing.TypeAlias = sqlm.Session


def create_engine(
    name: str,
    host: str,
    port: int,
    username: str | None = None,
    password: str | None = None,
    **kwargs,
) -> Engine:
    """Creates a SQL database engine."""

    if not username:
        username: str = env.get_value("PG_USERNAME")
    if not password:
        password: str = env.get_value("PG_PASSWORD")

    assert username is not None, "username not provided"
    assert password is not None, "password not provided"

    # NOTE: We only support postgresql databases for now
    url: str = f"postgresql+psycopg://{username}:{password}@{host}:{port}/{name}"
    engine: sqla.Engine = sqlm.create_engine(url, **kwargs)
    return engine


def verify_engine(engine: Engine) -> None | str:
    """Verifies if a database engine is able to connect."""
    try:
        connection: sqla.Connection = engine.connect()
        connection.close()
        return None
    except sqla.exc.SQLAlchemyError as error:
        return str(error)



"""Package with visualization functionality for plotting. The package uses
plotly as the plotting engine."""

from .geometry_plotting import (
    create_subplots,
    trace_registration_result,
)


__all__ = [
    "create_subplots",
    "trace_registration_result",
]



"""Module for Mynds record models."""

import textwrap
import typing

from enum import StrEnum, auto
from typing import ClassVar, Optional, TypeAlias

import numpy as np

from sqlmodel import SQLModel, Column, Field, Relationship
from sqlmodel import Integer, Float, ARRAY, JSON, Enum


class ChunkGroupRecord(SQLModel, table=True):
    """Class representing chunk group record."""

    id: Optional[int] = Field(default=None, primary_key=True)
    label: str
    chunks: list["ChunkRecord"] = Relationship(back_populates="group")


class ChunkRecord(SQLModel, table=True):
    """Class representing a chunk record."""

    id: Optional[int] = Field(default=None, primary_key=True)
    label: str
    meta: dict = Field(default_factory=dict, sa_type=JSON)

    # Create a one-to-many mapping from a chunk group to multiple chunks
    group_id: Optional[int] = Field(default=None, foreign_key="chunkgrouprecord.id")
    group: Optional["ChunkGroupRecord"] = Relationship(back_populates="chunks")

    sensors: list["SensorRecord"] = Relationship(
        back_populates="chunk",
        sa_relationship_kwargs={"lazy": "selectin"},
    )
    cameras: list["CameraRecord"] = Relationship(
        back_populates="chunk",
        sa_relationship_kwargs={"lazy": "selectin"},
    )

    stereo_rigs: list["StereoRigRecord"] = Relationship(
        back_populates="chunk",
        sa_relationship_kwargs={"lazy": "selectin"},
    )
    stereo_sensor_pairs: list["StereoSensorRecord"] = Relationship(
        back_populates="chunk",
        sa_relationship_kwargs={"lazy": "selectin"},
    )
    stereo_camera_pairs: list["StereoCameraPairRecord"] = Relationship(
        back_populates="chunk",
        sa_relationship_kwargs={"lazy": "selectin"},
    )


class CalibrationBaseRecord(SQLModel):
    """Class representing a calibration base record.

    For reference:
    https://docs.opencv.org/4.x/d9/d0c/group__calib3d.html
    """

    width: int
    height: int
    fx: float
    fy: float
    cx: float  # optical center as per OpenCV
    cy: float  # optical center as per OpenCV
    k1: float  # radial distortion as per OpenCV
    k2: float  # radial distortion as per OpenCV
    k3: float  # radial distortion as per OpenCV
    p1: float  # tangential distortion as per OpenCV
    p2: float  # tangential distortion as per OpenCV


class CalibrationRecord(CalibrationBaseRecord, table=True):
    """Class representing a calibration record."""

    id: Optional[int] = Field(default=None, primary_key=True)

    # Create a one-to-one mapping between a sensor and a calibration
    sensor: "SensorRecord" = Relationship(
        back_populates="calibration",
        sa_relationship_kwargs={"uselist": False, "lazy": "selectin"},
    )


class RectifiedCalibrationRecord(CalibrationBaseRecord, table=True):
    """Class representing a rectified calibration record."""

    id: Optional[int] = Field(default=None, primary_key=True)

    # Create a one-to-one mapping between a rectified sensor and calibration
    sensor: "RectifiedSensorRecord" = Relationship(
        back_populates="calibration",
        sa_relationship_kwargs={"uselist": False, "lazy": "selectin"},
    )


"""
Sensor records:
 - SensorBaseRecord
 - SensorRecord
 - RectifiedSensorRecord
"""


class SensorBaseRecord(SQLModel):
    """Class representing a sensor base record."""

    label: str
    width: int
    height: int

    location: Optional[list[float]] = Field(
        default=None,
        sa_type=ARRAY(Float, dimensions=1),
    )
    rotation: Optional[list[list[float]]] = Field(
        default=None,
        sa_type=ARRAY(Float, dimensions=2),
    )


class SensorRecord(SensorBaseRecord, table=True):
    """Class representing a sensor record."""

    id: Optional[int] = Field(default=None, primary_key=True)

    chunk_id: Optional[int] = Field(default=None, foreign_key="chunkrecord.id")
    chunk: Optional["ChunkRecord"] = Relationship(back_populates="sensors")

    calibration_id: Optional[int] = Field(
        default=None, foreign_key="calibrationrecord.id"
    )
    calibration: Optional["CalibrationRecord"] = Relationship(
        back_populates="sensor",
        sa_relationship_kwargs={"uselist": False},
    )

    rectified_sensor_id: Optional[int] = Field(
        default=None, foreign_key="sensorrecord.id"
    )
    rectified_sensor: Optional["RectifiedSensorRecord"] = Relationship(
        back_populates="source_sensor",
        sa_relationship_kwargs={"uselist": False},
    )


class RectifiedSensorRecord(SensorBaseRecord, table=True):
    """Class representing a rectified sensor record."""

    id: Optional[int] = Field(default=None, primary_key=True)

    calibration_id: Optional[int] = Field(
        default=None, foreign_key="rectifiedcalibrationrecord.id"
    )
    calibration: Optional["RectifiedCalibrationRecord"] = Relationship(
        back_populates="sensor",
        sa_relationship_kwargs={"uselist": False},
    )

    source_sensor_id: Optional[int] = Field(default=None, foreign_key="sensorrecord.id")
    source_sensor: Optional["SensorRecord"] = Relationship(
        back_populates="rectified_sensor",
        sa_relationship_kwargs={"uselist": False},
    )


"""
Reference models:
 - BaseReferenceRecord
 - PriorReferenceRecord
 - AlignedReferenceRecord
 - AssimilatedReferenceRecord
"""


class BaseReferenceRecord(SQLModel):
    """Class representing a base reference record."""

    epsg_code: int
    longitude: float
    latitude: float
    height: float
    yaw: float
    pitch: float
    roll: float


class PriorReferenceRecord(BaseReferenceRecord, table=True):
    """Class representing a prior camera reference record."""

    id: Optional[int] = Field(default=None, primary_key=True)

    camera_id: Optional[int] = Field(
        default=None, foreign_key="camerarecord.id", unique=True
    )
    camera: Optional["CameraRecord"] = Relationship(
        back_populates="prior_reference",
        sa_relationship_kwargs={"lazy": "select"},
    )


class AlignedReferenceRecord(BaseReferenceRecord, table=True):
    """Class representing an aligned camera reference record."""

    id: Optional[int] = Field(default=None, primary_key=True)

    camera_id: Optional[int] = Field(
        default=None, foreign_key="camerarecord.id", unique=True
    )
    camera: Optional["CameraRecord"] = Relationship(
        back_populates="aligned_reference",
        sa_relationship_kwargs={"lazy": "select"},
    )


class AssimilatedReferenceRecord(BaseReferenceRecord, table=True):
    """Class representing an assimilated camera reference record."""

    id: Optional[int] = Field(default=None, primary_key=True)

    camera_id: Optional[int] = Field(
        default=None, foreign_key="camerarecord.id", unique=True
    )
    camera: Optional["CameraRecord"] = Relationship(
        back_populates="assimilated_reference",
        sa_relationship_kwargs={"lazy": "select"},
    )


class CameraRecord(SQLModel, table=True):
    """Class representing a camera record."""

    id: Optional[int] = Field(default=None, primary_key=True)
    label: str
    image_label: str
    meta: dict[str, typing.Any] = Field(
        default_factory=dict,
        sa_type=JSON,
    )

    chunk_id: Optional[int] = Field(default=None, foreign_key="chunkrecord.id")
    chunk: Optional["ChunkRecord"] = Relationship(back_populates="cameras")

    # Every camera has a sensor
    sensor_id: Optional[int] = Field(default=None, foreign_key="sensorrecord.id")
    sensor: Optional["SensorRecord"] = Relationship(
        sa_relationship_kwargs={
            "uselist": False,
            "primaryjoin": lambda: CameraRecord.sensor_id == SensorRecord.id,
        }
    )

    prior_reference: Optional["PriorReferenceRecord"] = Relationship(
        back_populates="camera",
        sa_relationship_kwargs={"lazy": "select"},
    )
    aligned_reference: Optional["AlignedReferenceRecord"] = Relationship(
        back_populates="camera",
        sa_relationship_kwargs={"lazy": "select"},
    )
    assimilated_reference: Optional["AssimilatedReferenceRecord"] = Relationship(
        back_populates="camera",
        sa_relationship_kwargs={"lazy": "select"},
    )


class PixelMapRecord(SQLModel, table=True):
    """Class representing a pixel map record. The pixel map value is stored as
    a HxWx2 array with the X- and Y-coordinate mapping, respectively."""

    id: Optional[int] = Field(default=None, primary_key=True)
    values: list[list[list[float]]] = Field(sa_type=ARRAY(Float, dimensions=3))

    @property
    def height(self) -> int:
        """Returns the height of the pixel map."""
        return len(self.values)

    @property
    def width(self) -> int:
        """Returns the width of the pixel map."""
        return len(self.values[0])

    def to_array(self) -> np.ndarray:
        """Returns the pixel map as a numpy array."""
        return np.array(self.values).astype(np.float32)


"""
Stereo record models:
 - StereoSensorRecord
 - StereoCameraPairRecord
 - StereoPixelMapRecord
 - StereoRigRecord
"""


class StereoSensorComponentRecord(SQLModel, table=True):
    """Class representing a stereo sensor component."""

    id: Optional[int] = Field(default=None, primary_key=True)

    sensor_id: Optional[int] = Field(
        default=None, foreign_key="sensorrecord.id", unique=True
    )
    sensor: SensorRecord = Relationship(
        sa_relationship_kwargs={
            "primaryjoin": textwrap.dedent("""
                StereoSensorComponentRecord.sensor_id == SensorRecord.id
            """)
        }
    )

    pixel_map_id: Optional[int] = Field(
        default=None, foreign_key="pixelmaprecord.id", unique=True
    )
    pixel_map: PixelMapRecord = Relationship(
        sa_relationship_kwargs={
            "primaryjoin": textwrap.dedent("""
                StereoSensorComponentRecord.pixel_map_id == PixelMapRecord.id
            """)
        }
    )

    @property
    def rectified_sensor(self) -> RectifiedSensorRecord | None:
        """Gets the rectified sensor record."""
        return self.sensor.rectified_sensor

    @rectified_sensor.setter
    def rectified_sensor(self, sensor: RectifiedSensorRecord) -> None:
        """Sets the rectified sensor record."""
        self.sensor.rectified_sensor = sensor

    def is_rectified(self) -> bool:
        """Returns true if the sensor component is rectified."""
        return self.rectified_sensor is not None

    def has_pixel_map(self) -> bool:
        """Returns true if the sensor component has a pixel map."""
        return self.pixel_map is not None


class StereoSensorRecord(SQLModel, table=True):
    """Class representing a stereo sensor record. The stereo sensor combines
    two stereo sensor components, one master and one slave."""

    Component: ClassVar[TypeAlias] = StereoSensorComponentRecord

    id: Optional[int] = Field(default=None, primary_key=True)

    # A stereo sensor is uniquely identified by the two sensor ids
    master_component_id: Optional[int] = Field(
        default=None, foreign_key="stereosensorcomponentrecord.id", unique=True
    )
    master_component: StereoSensorComponentRecord = Relationship(
        sa_relationship_kwargs={
            "primaryjoin": textwrap.dedent("""
                StereoSensorRecord.master_component_id == StereoSensorComponentRecord.id
            """)
        }
    )

    slave_component_id: Optional[int] = Field(
        default=None, foreign_key="stereosensorcomponentrecord.id", unique=True
    )
    slave_component: StereoSensorComponentRecord = Relationship(
        sa_relationship_kwargs={
            "primaryjoin": textwrap.dedent("""
                StereoSensorRecord.slave_component_id == StereoSensorComponentRecord.id
            """)
        }
    )

    stereo_rig_id: Optional[int] = Field(
        default=None, foreign_key="stereorigrecord.id", unique=True
    )
    stereo_rig: Optional["StereoRigRecord"] = Relationship(
        sa_relationship_kwargs={
            "uselist": False,
        }
    )

    chunk_id: Optional[int] = Field(default=None, foreign_key="chunkrecord.id")
    chunk: Optional["ChunkRecord"] = Relationship(back_populates="stereo_sensor_pairs")

    @property
    def master(self) -> StereoSensorComponentRecord | None:
        """Gets the master stereo sensor component."""
        return self.master_component

    @master.setter
    def master(self, component: StereoSensorComponentRecord) -> None:
        """Sets the master stereo sensor component."""
        self.master_component = component

    @property
    def slave(self) -> StereoSensorComponentRecord | None:
        """Gets the slave stereo sensor component."""
        return self.slave_component

    @slave.setter
    def slave(self, component: StereoSensorComponentRecord) -> None:
        """Sets the slave stereo sensor component."""
        self.slave_component = component

    def is_rectified(self) -> bool:
        """Returns true if the stereo sensor components are rectified."""
        if self.master is None or self.slave is None:
            return False
        else:
            return self.master.is_rectified() and self.slave.is_rectified()

    def has_pixel_maps(self) -> bool:
        """Returns true if the stereo sensor has pixel maps."""
        return self.master.has_pixel_map() and self.slave.has_pixel_map()


class StereoCameraPairRecord(SQLModel, table=True):
    """Class representing a stereo camera pair record. A stereo camera links two
    camera records. The stereo camera is identified by the ids of the two
    cameras, and is owned by a chunk."""

    id: Optional[int] = Field(default=None, primary_key=True)

    master_camera_id: int = Field(foreign_key="camerarecord.id", unique=True)
    master_camera: CameraRecord = Relationship(
        sa_relationship_kwargs={
            "primaryjoin": lambda: StereoCameraPairRecord.master_camera_id
            == CameraRecord.id
        }
    )

    slave_camera_id: int = Field(foreign_key="camerarecord.id", unique=True)
    slave_camera: CameraRecord = Relationship(
        sa_relationship_kwargs={
            "primaryjoin": lambda: StereoCameraPairRecord.slave_camera_id
            == CameraRecord.id
        }
    )

    # TODO: Add validation that the two camera sensors are part of the sensor
    stereo_rig_id: Optional[int] = Field(default=None, foreign_key="stereorigrecord.id")
    stereo_rig: Optional["StereoRigRecord"] = Relationship(
        sa_relationship_kwargs={
            "uselist": False,
        }
    )

    chunk_id: Optional[int] = Field(default=None, foreign_key="chunkrecord.id")
    chunk: Optional["ChunkRecord"] = Relationship(back_populates="stereo_camera_pairs")

    @property
    def master(self) -> CameraRecord | None:
        """Gets the master camera."""
        return self.master_camera

    @master.setter
    def master(self, camera: CameraRecord) -> None:
        """Sets the master camera."""
        self.master_camera = camera

    @property
    def slave(self) -> CameraRecord | None:
        """Gets the slave camera."""
        return self.slave_camera

    @slave.setter
    def slave(self, camera: CameraRecord) -> None:
        """Sets the slave camera."""
        self.slave_camera = camera


class StereoRigRecord(SQLModel, table=True):
    """Class representing a stereo rig record. A stereo rig contains the sensor
    records for the physical sensor system, and the sensor records for the
    virtual/rectified sensor system, including the pixel map between the two
    sensor."""

    SensorComponent: ClassVar[TypeAlias] = StereoSensorComponentRecord
    SensorPair: ClassVar[TypeAlias] = StereoSensorRecord
    CameraPair: ClassVar[TypeAlias] = StereoCameraPairRecord
    PixelMap: ClassVar[TypeAlias] = PixelMapRecord

    id: Optional[int] = Field(default=None, primary_key=True)

    sensors: StereoSensorRecord = Relationship(
        back_populates="stereo_rig",
        sa_relationship_kwargs={
            "uselist": False,
        },
    )

    camera_pairs: list["StereoCameraPairRecord"] = Relationship(
        back_populates="stereo_rig"
    )

    chunk_id: Optional[int] = Field(default=None, foreign_key="chunkrecord.id")
    chunk: Optional["ChunkRecord"] = Relationship(back_populates="stereo_rigs")

    def is_rectified(self) -> bool:
        """Returns true if the stereo rig sensors are rectified."""
        return self.sensors.is_rectified()

    def has_pixel_maps(self) -> bool:
        """Returns true if the stereo rig has pixel maps."""
        return self.sensors.has_pixel_maps()



"""Package with visualization functionality for rendering 2D and 3D information.
The package is based on Open3D and OpenCV."""

from .geometry_renderers import (
    visualize_registration,
    visualize_registration_batch,
)


from .image_renderers import (
    WindowHandle,
    TrackbarData,
    create_window,
    render_image,
    destroy_window,
    destroy_all_windows,
    wait_key_input,
    StereoWindows,
    create_stereo_windows,
    render_stereo_geometry,
    create_stereo_geometry_color_image,
)


__all__ = [
    "visualize_registration",
    "visualize_registration_batch",
    "WindowHandle",
    "TrackbarData",
    "create_window",
    "render_image",
    "destroy_window",
    "destroy_all_windows",
    "wait_key_input",
    "StereoWindows",
    "create_stereo_windows",
    "render_stereo_geometry",
    "create_stereo_geometry_color_image",
]



"""Module for transforming locations between coordinate reference systems."""

import numpy as np
import pymap3d as pm


def _convert_location_geodetic2enu(
    latitude: float,
    longitude: float,
    height: float,
    *,
    center_latitude: float,
    center_longitude: float,
    center_height: float,
) -> tuple[float, float, float]:
    """Converts a row to ENU."""
    east, north, up = pm.enu.geodetic2enu(
        lat=latitude,
        lon=longitude,
        h=height,
        lat0=center_latitude,
        lon0=center_longitude,
        h0=center_height,
        deg=True,
    )
    return east, north, up


def convert_locations_geodetic2enu(
    locations: np.ndarray,
    *,
    center_longitude: float,
    center_latitude: float,
    center_height: float = 0.0,
) -> np.ndarray:
    """Converts locations from Geodetic (WGS84) to East-North-Up."""
    converted_locations: list[tuple] = [
        _convert_location_geodetic2enu(
            latitude=location[0],
            longitude=location[1],
            height=location[2],
            center_longitude=center_longitude,
            center_latitude=center_latitude,
            center_height=center_height,
        )
        for location in locations
    ]
    return np.array(converted_locations)


def _convert_location_enu2geodetic(
    east: float,
    north: float,
    up: float,
    center_longitude: float,
    center_latitude: float,
    center_height: float,
) -> tuple[float, float, float]:
    """Converts a row"""
    latitude, longitude, height = pm.enu.enu2geodetic(
        e=east,
        n=north,
        u=up,
        lat0=center_latitude,
        lon0=center_longitude,
        h0=center_height,
        deg=True,
    )
    return latitude, longitude, height


def convert_locations_enu2geodetic(
    locations: np.ndarray,
    *,
    center_longitude: float,
    center_latitude: float,
    center_height: float = 0.0,
) -> np.ndarray:
    """TODO"""
    converted_locations: list[tuple] = [
        _convert_location_enu2geodetic(
            east=location[0],
            north=location[1],
            up=location[2],
            center_longitude=center_longitude,
            center_latitude=center_latitude,
            center_height=center_height,
        )
        for location in locations
    ]
    return np.array(converted_locations)


def _convert_location_geodetic2ned(
    latitude: float,
    longitude: float,
    height: float,
    *,
    center_longitude: float,
    center_latitude: float,
    center_height: float = 0.0,
) -> tuple[float, float, float]:
    """Converts a location from geodetic to North-East-Down."""
    north, east, down = pm.ned.geodetic2ned(
        lat=latitude,
        lon=longitude,
        h=height,
        lat0=center_latitude,
        lon0=center_longitude,
        h0=center_height,
        deg=True,
    )
    return north, east, down


def convert_locations_geodetic2ned(
    locations: np.ndarray,
    *,
    center_longitude: float,
    center_latitude: float,
    center_height: float = 0.0,
) -> np.ndarray:
    """Converts locations from Geodetic (WGS84) to North-East-Down."""
    converted_locations: list[tuple] = [
        _convert_location_geodetic2ned(
            latitude=location[0],
            longitude=location[1],
            height=location[2],
            center_longitude=center_longitude,
            center_latitude=center_latitude,
            center_height=center_height,
        )
        for location in locations
    ]
    return np.array(converted_locations)


def _convert_location_ned2geodetic(
    north: float,
    east: float,
    down: float,
    *,
    center_longitude: float,
    center_latitude: float,
    center_height: float = 0.0,
) -> tuple[float, float, float]:
    """Converts a location from North-East-Down to Geodetic."""
    latitude, longitude, height = pm.ned.ned2geodetic(
        n=north,
        e=east,
        d=down,
        lat0=center_latitude,
        lon0=center_longitude,
        h0=center_height,
        deg=True,
    )
    return latitude, longitude, height


def convert_locations_ned2geodetic(
    locations: np.ndarray,
    *,
    center_longitude: float,
    center_latitude: float,
    center_height: float = 0.0,
) -> np.ndarray:
    """Converts locations from NED to Geodetic.

    :arg locations: Nx3 array of [north, east, down]
    :return: Nx3 array of [latitude, longitude, height]
    """
    converted_locations: list[tuple] = [
        _convert_location_ned2geodetic(
            north=location[0],
            east=location[1],
            down=location[2],
            center_longitude=center_longitude,
            center_latitude=center_latitude,
            center_height=center_height,
        )
        for location in locations
    ]
    return np.array(converted_locations)



"""Package with camera export functionality."""

from .export import (
    export_cameras,
    export_stereo_cameras,
    tabularize_cameras,
    tabularize_stereo_cameras,
)

__all__ = [
    "export_cameras",
    "export_stereo_cameras",
    "tabularize_cameras",
    "tabularize_stereo_cameras",
]



"""Module for full registration methods."""

from typing import TypeVar

import numpy as np
import open3d

import open3d.pipelines.registration as reg

from .types import RegistrationResult


Key: TypeVar = TypeVar("Key")


def build_pose_graph(
    results: dict[Key, dict[Key, RegistrationResult]],
) -> reg.PoseGraph:
    """Builds a pose graph from registered point clouds."""

    odometry: np.ndarray = np.identity(4)

    pose_graph: reg.PoseGraph = reg.PoseGraph()
    pose_graph.nodes.append(reg.PoseGraphNode(odometry))

    for source_id, registrations in results.items():
        for target_id, result in registrations.items():
            if target_id == source_id + 1:  # odometry case
                odometry: np.ndarray = np.dot(result.transformation, odometry)

                pose_graph.nodes.append(
                    reg.PoseGraphNode(
                        np.linalg.inv(odometry),
                    )
                )
                pose_graph.edges.append(
                    reg.PoseGraphEdge(
                        source_id,
                        target_id,
                        result.transformation,
                        result.information,
                        uncertain=False,
                    )
                )
            else:  # loop closure case
                pose_graph.edges.append(
                    reg.PoseGraphEdge(
                        source_id,
                        target_id,
                        result.transformation,
                        result.information,
                        uncertain=True,
                    )
                )

    return pose_graph


def optimize_pose_graph(
    pose_graph: reg.PoseGraph,
    correspondence_distance: float,
    prune_threshold: float,
    preference_loop_closure: float,
    reference_node: int = -1,
) -> reg.PoseGraph:
    """Optimizes a pose graph by optimizing and pruning graph edges."""

    method = reg.GlobalOptimizationLevenbergMarquardt()

    criteria = reg.GlobalOptimizationConvergenceCriteria()

    option = reg.GlobalOptimizationOption(
        max_correspondence_distance=correspondence_distance,
        edge_prune_threshold=prune_threshold,
        preference_loop_closure=preference_loop_closure,
        reference_node=reference_node,
    )

    reg.global_optimization(
        pose_graph,
        method,
        criteria,
        option,
    )

    return pose_graph



"""Module for common image processors."""

from typing import Literal

import cv2
import numpy as np

from .image_types import Image, PixelFormat


# Create a literal type to enable type annotations
ColorConversionCode = Literal[
    cv2.COLOR_GRAY2RGB,
    cv2.COLOR_GRAY2BGR,
    cv2.COLOR_GRAY2RGBA,
    cv2.COLOR_GRAY2BGRA,
    # GRAY -> RGB / BGR / RGBA / BGRA
    cv2.COLOR_RGB2GRAY,
    cv2.COLOR_RGB2BGR,
    cv2.COLOR_RGB2RGBA,
    cv2.COLOR_RGB2BGRA,
    # RGB -> GRAY / BGR / RGBA / BGRA
    cv2.COLOR_BGR2GRAY,
    cv2.COLOR_BGR2RGB,
    cv2.COLOR_BGR2RGBA,
    cv2.COLOR_BGR2BGRA,
    # BGR -> GRAY / RGB / RGBA / BGRA
    cv2.COLOR_RGBA2GRAY,
    cv2.COLOR_RGBA2RGB,
    cv2.COLOR_RGBA2BGR,
    cv2.COLOR_RGBA2BGRA,
    # RGBA -> GRAY / RGB / BGR / RGBA
    cv2.COLOR_BGRA2GRAY,
    cv2.COLOR_BGRA2RGB,
    cv2.COLOR_BGRA2BGR,
    cv2.COLOR_BGRA2RGBA,
]


def convert_image_color(image: Image, to_format: PixelFormat) -> Image:
    """Converts an image to the desired format."""
    from_format: PixelFormat = image.pixel_format
    conversion: ColorConversionCode = _map_formats_to_conversion(from_format, to_format)
    converted: np.ndarray = cv2.cvtColor(image.to_array(), conversion)
    return Image.from_array(converted, to_format)


def _map_formats_to_conversion(
    from_format: PixelFormat, to_format: PixelFormat
) -> ColorConversionCode:
    """Maps source and target pixel formats to a color convertsion function."""
    match from_format:
        case PixelFormat.GRAY:
            return _get_color_conversion_gray(to_format)
        case PixelFormat.RGB:
            return _get_color_conversion_rgb(to_format)
        case PixelFormat.BGR:
            return _get_color_conversion_bgr(to_format)
        case PixelFormat.RGBA:
            return _get_color_conversion_rgba(to_format)
        case PixelFormat.BGRA:
            return _get_color_conversion_bgra(to_format)
        case _:
            raise NotImplementedError(f"invalid pixel format: {from_format}")


def _get_color_conversion_gray(to_format: PixelFormat) -> ColorConversionCode:
    """Gets the color conversion format for grayscale images."""
    match to_format:
        case PixelFormat.RGB:
            return cv2.COLOR_GRAY2RGB
        case PixelFormat.BGR:
            return cv2.COLOR_GRAY2BGR
        case PixelFormat.RGBA:
            return cv2.COLOR_GRAY2RGBA
        case PixelFormat.BGRA:
            return cv2.COLOR_GRAY2BGRA
        case _:
            raise NotImplementedError(f"invalid target format: {to_format}")


def _get_color_conversion_rgb(to_format: PixelFormat) -> ColorConversionCode:
    """Gets the color conversion format for RGB images."""
    match to_format:
        case PixelFormat.GRAY:
            return cv2.COLOR_RGB2GRAY
        case PixelFormat.BGR:
            return cv2.COLOR_RGB2BGR
        case PixelFormat.RGBA:
            return cv2.COLOR_RGB2RGBA
        case PixelFormat.BGRA:
            return cv2.COLOR_RGB2BGRA
        case _:
            raise NotImplementedError(f"invalid target format: {to_format}")


def _get_color_conversion_bgr(to_format: PixelFormat) -> ColorConversionCode:
    """Gets the color conversion format for BGR images."""
    match to_format:
        case PixelFormat.GRAY:
            return cv2.COLOR_BGR2GRAY
        case PixelFormat.RGB:
            return cv2.COLOR_BGR2RGB
        case PixelFormat.RGBA:
            return cv2.COLOR_BGR2RGBA
        case PixelFormat.BGRA:
            return cv2.COLOR_BGR2BGRA
        case _:
            raise NotImplementedError(f"invalid target format: {to_format}")


def _get_color_conversion_rgba(to_format: PixelFormat) -> ColorConversionCode:
    """Gets the color conversion format for RGBA images."""
    match to_format:
        case PixelFormat.GRAY:
            return cv2.COLOR_RGBA2GRAY
        case PixelFormat.RGB:
            return cv2.COLOR_RGBA2RGB
        case PixelFormat.BGR:
            return cv2.COLOR_RGBA2BGR
        case PixelFormat.BGRA:
            return cv2.COLOR_RGBA2BGRA
        case _:
            raise NotImplementedError(f"invalid color format: {to_format}")


def _get_color_conversion_bgra(to_format: PixelFormat) -> ColorConversionCode:
    """Gets the color conversion format for BGRA images."""
    match to_format:
        case PixelFormat.GRAY:
            return cv2.COLOR_BGRA2GRAY
        case PixelFormat.RGB:
            return cv2.COLOR_BGRA2RGB
        case PixelFormat.BGR:
            return cv2.COLOR_BGRA2BGR
        case PixelFormat.RGBA:
            return cv2.COLOR_BGRA2RGBA
        case _:
            raise NotImplementedError(f"invalid color format: {to_format}")


def convert_image_to_rgb(image: Image) -> Image:
    """Converts an image to RGB."""
    match image.pixel_format:
        case PixelFormat.GRAY:
            return _convert_grayscale_to_rgb(image)
        case PixelFormat.RGB:
            return image
        case _:
            raise NotImplementedError("invalid image pixel format")


def _convert_grayscale_to_rgb(image: Image) -> Image:
    """Converts a grayscale image to RGB."""
    values: np.ndarray = cv2.cvtColor(image.to_array(), cv2.COLOR_GRAY2RGB)
    return Image.from_array(values, PixelFormat.RGB)


def apply_color_map(image: Image) -> Image:
    """Applies a color map to the image values."""
    values: np.ndarray = cv2.applyColorMap(image.to_array(), cv2.COLORMAP_JET)
    return Image.from_array(values, PixelFormat.RGB)



"""Module for exporting camera references."""

from pathlib import Path
from typing import Any

import polars as pl

import mynd.schemas as schemas

from mynd.config import write_config
from mynd.utils.log import logger


def export_cameras(path: Path, cameras: list[schemas.CameraSchema]) -> Path:
    """Export a collection of cameras to file."""

    camera_data: list[dict] = [camera.model_dump() for camera in cameras]
    result: Path | str = write_config(path, camera_data)

    if isinstance(result, str):
        logger.error(result)
        raise IOError(result)

    return result


def export_stereo_cameras(
    path: Path, stereo_cameras: list[schemas.StereoCameraPairSchema]
) -> Path:
    """Exports a collection of stereo cameras to file."""

    stereo_camera_data: list[dict] = [
        stereo_camera.model_dump() for stereo_camera in stereo_cameras
    ]
    result: Path | str = write_config(path, stereo_camera_data)

    if isinstance(result, str):
        logger.error(result)
        raise IOError(result)

    return result


def tabularize_cameras(cameras: list[schemas.CameraSchema]) -> pl.DataFrame:
    """Tabularizes a collection of cameras."""

    rows: list[dict[str, Any]] = list()
    for camera in cameras:
        row: dict[str, Any] = {
            "camera_id": camera.id,
            "camera_label": camera.label,
            "image_label": camera.image_label,
            "readings": camera.readings,
            "sensor_id": camera.sensor.id,
        }

        if camera.interpolated_reference is not None:
            row.update(
                {
                    "latitude": camera.interpolated_reference.latitude,
                    "longitude": camera.interpolated_reference.longitude,
                    "height": camera.interpolated_reference.height,
                    "yaw": camera.interpolated_reference.yaw,
                    "pitch": camera.interpolated_reference.pitch,
                    "roll": camera.interpolated_reference.roll,
                }
            )

        rows.append(row)

    return pl.from_dicts(rows)


def tabularize_stereo_cameras(
    stereo_cameras: list[schemas.StereoCameraPairSchema],
) -> pl.DataFrame:
    """Tabularizes a collection of stereo cameras."""

    rows: list[dict] = list()
    for stereo_camera in stereo_cameras:
        assert (
            stereo_camera.master.interpolated_reference is not None
        ), "missing interpolated reference for master camera"
        common_fields: dict[str, Any] = _get_stereo_common_fields(stereo_camera)
        master_fields: dict[str, Any] = _get_camera_fields(stereo_camera.master)
        slave_fields: dict[str, Any] = _get_camera_fields(stereo_camera.slave)

        reference_fields: dict[str, Any] = _get_reference_fields(
            stereo_camera.master.interpolated_reference
        )

        master_fields: dict[str, Any] = {
            f"master.{key}": value for key, value in master_fields.items()
        }
        slave_fields: dict[str, Any] = {
            f"slave.{key}": value for key, value in slave_fields.items()
        }

        row: dict = common_fields | master_fields | slave_fields | reference_fields
        rows.append(row)

    return pl.from_dicts(rows)


def _get_stereo_common_fields(
    stereo_camera: schemas.StereoCameraPairSchema,
) -> dict[str, Any]:
    """Retrieves common fields for a stereo camera."""
    return {
        "stereo_camera_id": stereo_camera.id,
        "stereo_rig_id": stereo_camera.stereo_rig.id,
    }


def _get_camera_fields(camera: schemas.CameraSchema) -> dict[str, Any]:
    """Retrieves selected fields from a stereo camera master."""
    return {
        "camera_id": camera.id,
        "camera_label": camera.label,
        "image_label": camera.image_label,
        "readings": camera.readings,
    }


def _get_reference_fields(
    reference: schemas.CameraSchema.ReferenceType,
) -> dict[str, Any]:
    """Retrieves fields from a camera reference."""
    return {
        "latitude": reference.latitude,
        "longitude": reference.longitude,
        "height": reference.height,
        "yaw": reference.yaw,
        "pitch": reference.pitch,
        "roll": reference.roll,
    }



"""Module for functionality related to range maps. Currently the module includes functionality
for computing range from disparity, 3D points from range, and normal maps from range maps."""

import warnings

import kornia
import kornia.geometry.depth as kgd
import numpy as np
import torch

from mynd.image import Image, PixelFormat


warnings.warn = lambda *args, **kwargs: None


def compute_ranges_from_disparities(
    disparity: Image,
    baseline: float,
    focal_length: float,
) -> Image:
    """Computes a range map from the given disparity map and camera matrix. Returns the
    range map as a HxW array with float32 values."""

    assert (
        disparity.pixel_format == PixelFormat.X
    ), f"invalid disparity map format: {disparity.pixel_format}"

    # Kornia expects a disparity tensor of shape Bx1xHxW
    range_tensor: torch.Tensor = kgd.depth_from_disparity(
        disparity=_disparity_map_to_tensor(disparity.to_array()),
        baseline=baseline,
        focal=focal_length,
    )

    range_values: np.ndarray = _range_tensor_to_numpy(range_tensor)

    return Image.from_array(range_values, PixelFormat.X)


def compute_points_from_ranges(
    range_map: Image,
    camera_matrix: np.ndarray,
    normalize_points: bool = False,
) -> Image:
    """Computes 3D points based on the given range map and camera matrix. Returns
    the points as a HxWx3 array with float32 values."""

    assert (
        range_map.pixel_format == PixelFormat.X
    ), f"invalid range map format: {range_map.pixel_format}"

    # Kornia expects a range tensor of shape Bx1xHxW
    points_tensor: torch.Tensor = kgd.depth_to_3d_v2(
        depth=_range_map_to_tensor(range_map.to_array()),
        camera_matrix=_camera_matrix_to_tensor(camera_matrix),
        normalize_points=normalize_points,
    )

    # Convert 3D points back to XYZ format
    point_values: np.ndarray = _point_tensor_to_numpy(points_tensor)

    return Image.from_array(point_values, PixelFormat.XYZ)


def compute_normals_from_ranges(
    range_map: Image,
    camera_matrix: np.ndarray,
    flipped: bool = False,
    normalize_points: bool = False,
) -> Image:
    """Computes normal map based on the given range map and camera matrix. Returns
    the normals as float32 unit vectors. If flipped is true, the normals are defined
    with positive x-, y-, and z pointing right, down, and away as seen by the camera.
    """

    assert (
        range_map.pixel_format == PixelFormat.X
    ), f"invalid range map format: {range_map.pixel_format}"

    normal_tensor: torch.Tensor = kgd.depth_to_normals(
        depth=_range_map_to_tensor(range_map.to_array()),
        camera_matrix=_camera_matrix_to_tensor(camera_matrix),
        normalize_points=normalize_points,
    )

    normals: np.ndarray = _normal_tensor_to_numpy(normal_tensor)
    normals: np.ndarray = _normalize_normal_map(normals)

    if flipped:
        normals: np.ndarray = -normals

    # TODO: Convert to int8
    normals: np.ndarray = _convert_normal_map_to_int8(normals)

    return Image.from_array(normals, PixelFormat.XYZ)


def fill_range_map_dilation(range_map: Image) -> Image:
    """Fills a range map by dilation."""

    assert isinstance(range_map, Image), "invalid range map type"

    range_values: np.ndarray = np.squeeze(range_map.to_array())

    RANGE_LOWER: float = 0.1
    mask: np.ndarray = range_values < RANGE_LOWER
    KERNEL: torch.Tensor = torch.ones(51, 51)

    # TODO: Downscale range maps to increase computational efficiency
    range_tensor: torch.Tensor = _range_map_to_tensor(range_values)
    dilated_range_tensor: torch.Tensor = kornia.morphology.dilation(
        range_tensor, KERNEL, engine="unfold"
    )
    dilated_range_array: np.ndarray = _range_tensor_to_numpy(dilated_range_tensor)

    range_values[mask] = dilated_range_array[mask]

    return Image.from_array(range_values, pixel_format=range_map.pixel_format)


def _camera_matrix_to_tensor(camera_matrix: np.ndarray) -> torch.Tensor:
    """Converts a 3x3 camera matrix into a 1x3x3 torch tensor."""
    assert camera_matrix.shape == (
        3,
        3,
    ), f"invalid camera matrix shape: {camera_matrix.shape}"
    camera_matrix: np.ndarray = np.squeeze(camera_matrix)
    return torch.from_numpy(camera_matrix.copy()).view(1, 3, 3)


def _normalize_normal_map(normals: np.ndarray) -> np.ndarray:
    """Normalizes normals by their L2 norm."""
    norms: np.ndarray = np.linalg.norm(normals, axis=2)

    NORM_THRESHOLD: float = 0.0000001
    invalid: np.ndarray = norms < NORM_THRESHOLD
    norms[invalid] = 1.0
    normals[invalid] = np.zeros(3)

    # Convert normals into unit vectors
    normals /= norms[:, :, np.newaxis]
    return normals


def _convert_normal_map_to_int8(normals: np.ndarray) -> np.ndarray:
    """Converts a normal map from float32 unit vectors to int8 data type."""
    scale: int = np.iinfo(np.int8).max
    return np.round(scale * normals).astype(np.int8)


def _range_map_to_tensor(range_map: np.ndarray) -> torch.Tensor:
    """Converts a HxWx1 range map into a 1x1xHxW torch tensor."""
    range_map: np.ndarray = np.squeeze(range_map)
    return torch.from_numpy(range_map.copy()).view(
        1, 1, range_map.shape[0], range_map.shape[1]
    )


def _disparity_map_to_tensor(disparity: np.ndarray) -> torch.Tensor:
    """Converts a HxW disparity map into a 1x1xHxW torch tensor."""
    disparity: np.ndarray = np.squeeze(disparity)
    return torch.from_numpy(disparity.copy()).view(
        1, 1, disparity.shape[0], disparity.shape[1]
    )


def _range_map_to_tensor(range_map: np.ndarray) -> torch.Tensor:
    """Converts a HxW range map into a 1x1xHxW torch tensor."""
    range_map: np.ndarray = np.squeeze(range_map)
    return torch.from_numpy(range_map.copy()).view(
        1, 1, range_map.shape[0], range_map.shape[1]
    )


def _range_tensor_to_numpy(range_tensor: torch.Tensor) -> np.ndarray:
    """Converts a 1x1xHxW range map into a HxWx1 array."""
    assert (
        range_tensor.shape[0] == 1
    ), f"invalid range tensor shape - {range_tensor.shape}"
    assert (
        range_tensor.shape[1] == 1
    ), f"invalid range tensor shape - {range_tensor.shape}"
    return np.squeeze(range_tensor.numpy()).astype(np.float32)


def _point_tensor_to_numpy(point_tensor: torch.Tensor) -> np.ndarray:
    """Converts a 1x3xHxW point tensor into a HxWx3 array"""
    assert (
        point_tensor.shape[0] == 1
    ), f"invalid point tensor shape - {point_tensor.shape}"
    assert (
        point_tensor.shape[1] == 3
    ), f"invalid point tensor shape - {point_tensor.shape}"
    return np.squeeze(point_tensor.numpy()).transpose((1, 2, 0)).astype(np.float32)


def _normal_tensor_to_numpy(normal_tensor: torch.Tensor) -> np.ndarray:
    """Converts a 1x3xHxW points tensor into a HxWx3 array"""
    assert (
        normal_tensor.shape[0] == 1
    ), f"invalid points tensor shape - {normal_tensor.shape}"
    assert (
        normal_tensor.shape[1] == 3
    ), f"invalid points tensor shape - {normal_tensor.shape}"
    return np.squeeze(normal_tensor.numpy()).transpose((1, 2, 0)).astype(np.float32)



"""Module for point cloud processors, i.e. including filters for spacing and confidence."""

import copy

import numpy as np
import open3d

from mynd.geometry.point_cloud import PointCloud

import open3d.pipelines.registration as reg

from .types import (
    Feature,
    RegistrationResult,
    FeatureExtractor,
    FeatureMatcher,
    PointCloudAligner,
)


"""
Worker functions:
 - extract_fpfh_features
 - register_features_fast
 - register_features_ransac
"""


def extract_fpfh_features(input: PointCloud, radius: float, neighbours: int) -> Feature:
    """Compute fast point feature histrograms (FPFH)"""
    return reg.compute_fpfh_feature(
        input=input,
        search_param=open3d.geometry.KDTreeSearchParamHybrid(
            radius=radius,
            max_nn=neighbours,
        ),
    )


def register_features_fast(
    source: PointCloud,
    target: PointCloud,
    feature_extractor: FeatureExtractor,
    distance_threshold: float,
) -> RegistrationResult:
    """Extracts features and performs registration with FAST matching."""

    source: PointCloud = copy.deepcopy(source)
    target: PointCloud = copy.deepcopy(target)

    source_features: Feature = feature_extractor(input=source)
    target_features: Feature = feature_extractor(input=target)

    return _match_features_fast(
        source=source,
        target=target,
        source_features=source_features,
        target_features=target_features,
        distance_threshold=distance_threshold,
    )


def register_features_ransac(
    source: PointCloud,
    target: PointCloud,
    feature_extractor: FeatureExtractor,
    estimation_method: reg.TransformationEstimation,
    validators: list[reg.CorrespondenceChecker],
    convergence_criteria: reg.RANSACConvergenceCriteria,
    distance_threshold: float,
    sample_count: int = 3,
    mutual_filter: bool = True,
) -> RegistrationResult:
    """Extracts features and performs registration with RANSAC matching."""

    source: PointCloud = copy.deepcopy(source)
    target: PointCloud = copy.deepcopy(target)

    source_features: Feature = feature_extractor(input=source)
    target_features: Feature = feature_extractor(input=target)

    return _match_features_ransac(
        source=source,
        target=target,
        source_features=source_features,
        target_features=target_features,
        estimation_method=estimation_method,
        validators=validators,
        convergence_criteria=convergence_criteria,
        distance_threshold=distance_threshold,
        sample_count=sample_count,
        mutual_filter=mutual_filter,
    )


"""
Factories functions:
 - create_correspondence_validators
 - create_fpfh_extractor
 - create_convergence_criteria_ransac
 - create_ransac_registrator
 - create_point_to_point_estimator
 - create_point_to_plane_estimator
"""


def create_correspondence_validators(
    distance_threshold: float,
    edge_threshold: float | None = None,
    normal_threshold: float | None = None,
) -> list[reg.CorrespondenceChecker]:
    """Generates a set of correspondence validators."""

    validators: list = [reg.CorrespondenceCheckerBasedOnDistance(distance_threshold)]

    if edge_threshold:
        validators.append(reg.CorrespondenceCheckerBasedOnEdgeLength(edge_threshold))
    if normal_threshold:
        validators.append(reg.CorrespondenceCheckerBasedOnNormal(normal_threshold))

    return validators


def create_fpfh_extractor(radius: float, neighbours: int) -> FeatureExtractor:
    """Creates a FPFH feature extractor."""

    def feature_extractor_wrapper(input: PointCloud) -> Feature:
        """Wraps a FPFH feature extractor."""
        return extract_fpfh_features(input=input, radius=radius, neighbours=neighbours)

    return feature_extractor_wrapper


def create_convergence_criteria_ransac(
    max_iteration: int,
    confidence: float,
) -> reg.RANSACConvergenceCriteria:
    """Creates a RANSAC convergence convergence criteria."""
    return reg.RANSACConvergenceCriteria(
        max_iteration=max_iteration, confidence=confidence
    )


def create_ransac_registrator(
    feature_extractor: FeatureExtractor,
    estimation_method: reg.TransformationEstimation,
    validators: list[reg.CorrespondenceChecker],
    convergence_criteria: reg.RANSACConvergenceCriteria,
    distance_threshold: float,
    sample_count: int = 3,
    mutual_filter: bool = True,
) -> PointCloudAligner:
    """Creates a wrapper around a RANSAC registrator."""

    def ransac_registrator_wrapper(
        source: PointCloud,
        target: PointCloud,
    ) -> RegistrationResult:
        """Wraps a RANSAC feature registrator."""

        return register_features_ransac(
            source,
            target,
            feature_extractor=feature_extractor,
            estimation_method=estimation_method,
            validators=validators,
            convergence_criteria=convergence_criteria,
            distance_threshold=distance_threshold,
            sample_count=sample_count,
            mutual_filter=mutual_filter,
        )

    return ransac_registrator_wrapper


def create_point_to_point_estimator(
    with_scaling: bool = False,
) -> reg.TransformationEstimation:
    """Creates a point to point transformation estimator."""
    return reg.TransformationEstimationPointToPoint(with_scaling=with_scaling)


def create_point_to_plane_estimator(
    kernel: reg.RobustKernel | None = None,
) -> reg.TransformationEstimation:
    """Creates a point to plane transformation estimator."""
    return reg.TransformationEstimationPointToPlane(kernel=kernel)


"""
Helpers functions:
 - _extract_features_and_register
 - _match_features_fast
 - _match_features_ransac
"""


def _extract_features_and_register(
    source: PointCloud,
    target: PointCloud,
    feature_extractor: FeatureExtractor,
    feature_matcher: FeatureMatcher,
) -> RegistrationResult:
    """Executor for feature-based registration methods."""

    source_features: Feature = feature_extractor(input=source)
    target_features: Feature = feature_extractor(input=target)

    result: RegistrationResult = feature_matcher(
        source=source,
        target=target,
        source_features=source_features,
        target_features=target_features,
    )

    return result


def _match_features_fast(
    source: PointCloud,
    target: PointCloud,
    source_features: Feature,
    target_features: Feature,
    distance_threshold: float,
) -> RegistrationResult:
    """Wrapper function for Open3D feature based FAST registration method."""

    option: reg.FastGlobalRegistrationOption = reg.FastGlobalRegistrationOption(
        maximum_correspondence_distance=distance_threshold,
    )

    result: reg.RegistrationResult = reg.registration_fgr_based_on_feature_matching(
        source=source,
        target=target,
        source_feature=source_features,
        target_feature=target_features,
        option=option,
    )

    information_matrix: np.ndarray = reg.get_information_matrix_from_point_clouds(
        source=source,
        target=target,
        max_correspondence_distance=distance_threshold,
        transformation=result.transformation,
    )

    return RegistrationResult(
        fitness=result.fitness,
        inlier_rmse=result.inlier_rmse,
        correspondence_set=np.asarray(result.correspondence_set),
        transformation=result.transformation,
        information=information_matrix,
    )


def _match_features_ransac(
    source: PointCloud,
    target: PointCloud,
    source_features: Feature,
    target_features: Feature,
    estimation_method: reg.TransformationEstimation,
    validators: list[reg.CorrespondenceChecker],
    convergence_criteria: reg.RANSACConvergenceCriteria,
    distance_threshold: float,
    sample_count: int = 3,
    mutual_filter: bool = True,
) -> RegistrationResult:
    """Wrapper function for Open3D feature based RANSAC registration method."""

    result: reg.RegistrationResult = reg.registration_ransac_based_on_feature_matching(
        source=source,
        target=target,
        source_feature=source_features,
        target_feature=target_features,
        max_correspondence_distance=distance_threshold,
        mutual_filter=mutual_filter,
        ransac_n=sample_count,
        estimation_method=estimation_method,
        checkers=validators,
        criteria=convergence_criteria,
    )

    information_matrix: np.ndarray = reg.get_information_matrix_from_point_clouds(
        source=source,
        target=target,
        max_correspondence_distance=distance_threshold,
        transformation=result.transformation,
    )

    return RegistrationResult(
        fitness=result.fitness,
        inlier_rmse=result.inlier_rmse,
        correspondence_set=np.asarray(result.correspondence_set),
        transformation=result.transformation,
        information=information_matrix,
    )



"""Module for estimating stereo geometry, i.e. estimating ranges and normals from image pairs."""

from dataclasses import dataclass
from typing import Callable

from mynd.schemas import StereoSensorPairSchema
from mynd.image import Image, PixelFormat
from mynd.utils.containers import Pair

from .pixel_map import PixelMap, remap_image_pixels
from .range_map import compute_ranges_from_disparities, compute_normals_from_ranges
from .stereo_matcher import StereoMatcher


@dataclass
class StereoGeometry:
    """Class representing a stereo geometry."""

    rectified_images: Pair[Image]
    disparities: Pair[Image]
    range_maps: Pair[Image] | None = None
    normal_maps: Pair[Image] | None = None


ImageFilter = Callable[[Image], Image]
DisparityFilter = Callable[[Image], Image]


def compute_stereo_geometry(
    sensors_rectified: StereoSensorPairSchema,
    images_rectified: Pair[Image],
    matcher: StereoMatcher,
    image_filter: ImageFilter | None = None,
    disparity_filter: DisparityFilter | None = None,
) -> StereoGeometry:
    """Computes range and normal maps for a rectified stereo setup, a disparity matcher, and
    a pair of images."""

    if image_filter:
        filtered_images: Pair[Image] = Pair(
            first=image_filter(images_rectified.first),
            second=image_filter(images_rectified.second),
        )
    else:
        filtered_images: Pair[Image] = images_rectified

    # Estimate disparity from rectified images
    disparity_maps: Pair[Image] = matcher(
        left=filtered_images.first,
        right=filtered_images.second,
    )

    if disparity_filter:
        disparity_maps: Pair[Image] = Pair(
            first=disparity_filter(disparity_maps.first),
            second=disparity_filter(disparity_maps.second),
        )

    # Estimate range from disparity
    range_maps: Pair[Image] = Pair(
        first=compute_ranges_from_disparities(
            disparity=disparity_maps.first,
            baseline=sensors_rectified.baseline,
            focal_length=sensors_rectified.master.calibration.focal_length,
        ),
        second=compute_ranges_from_disparities(
            disparity=disparity_maps.second,
            baseline=sensors_rectified.baseline,
            focal_length=sensors_rectified.slave.calibration.focal_length,
        ),
    )

    # Estimate normals from range maps
    normal_maps: Pair[Image] = Pair(
        first=compute_normals_from_ranges(
            range_map=range_maps.first,
            camera_matrix=sensors_rectified.master.calibration.projection_matrix,
            flipped=True,
        ),
        second=compute_normals_from_ranges(
            range_map=range_maps.second,
            camera_matrix=sensors_rectified.slave.calibration.projection_matrix,
            flipped=True,
        ),
    )

    return StereoGeometry(
        rectified_images=images_rectified,
        disparities=disparity_maps,
        range_maps=range_maps,
        normal_maps=normal_maps,
    )


def compute_stereo_range_maps(
    rectified_sensors: StereoSensorPairSchema,
    rectified_images: Pair[Image],
    matcher: StereoMatcher,
    disparity_filter: DisparityFilter | None = None,
) -> Pair[Image]:
    """Computes range maps for a rectified stereo setup."""

    disparity_maps: Pair[Image] = matcher(
        left=rectified_images.first,
        right=rectified_images.second,
    )

    if disparity_filter:
        disparity_maps: Pair[Image] = Pair(
            first=disparity_filter(disparity_maps.first),
            second=disparity_filter(disparity_maps.second),
        )

    master_ranges: Image = compute_ranges_from_disparities(
        disparity=disparity_maps.first,
        baseline=rectified_sensors.baseline,
        focal_length=rectified_sensors.master.calibration.focal_length,
    )

    slave_ranges: Image = compute_ranges_from_disparities(
        disparity=disparity_maps.second,
        baseline=rectified_sensors.baseline,
        focal_length=rectified_sensors.slave.calibration.focal_length,
    )

    return Pair(master_ranges, slave_ranges)


def compute_stereo_normal_maps(
    rectified_sensors: StereoSensorPairSchema,
    range_maps: Pair[Image],
) -> Pair[Image]:
    """Computes stereo normal maps from a pair of rectified sensors and range
    maps."""

    first_normals: Image = compute_normals_from_ranges(
        range_map=range_maps.first,
        camera_matrix=rectified_sensors.master.calibration.projection_matrix,
        flipped=True,
    )

    second_normals: Image = compute_normals_from_ranges(
        range_map=range_maps.second,
        camera_matrix=rectified_sensors.slave.calibration.projection_matrix,
        flipped=True,
    )

    return Pair(first_normals, second_normals)


def distort_stereo_geometry(
    inverse_pixel_maps: Pair[PixelMap],
    range_maps: Pair[Image],
    normal_maps: Pair[Image],
) -> tuple[Pair[Image], ...]:
    """Distorts range and normal maps for the given stereo geometry."""

    distorted_ranges: Pair[Image] = Pair(
        first=remap_image_pixels(
            image=range_maps.first,
            pixel_map=inverse_pixel_maps.first,
        ),
        second=remap_image_pixels(
            image=range_maps.second,
            pixel_map=inverse_pixel_maps.second,
        ),
    )

    distorted_normals: Pair[Image] = Pair(
        first=remap_image_pixels(
            image=normal_maps.first,
            pixel_map=inverse_pixel_maps.first,
        ),
        second=remap_image_pixels(
            image=normal_maps.second,
            pixel_map=inverse_pixel_maps.second,
        ),
    )

    return distorted_ranges, distorted_normals



"""Module for common geometry transformations."""

import numpy as np

from scipy.spatial.transform import Rotation


def decompose_rigid_transform(transformation: np.ndarray) -> tuple:
    """Decomposes a 3D rigid body transformation into scale, rotation, and translation."""

    assert transformation.shape == (
        4,
        4,
    ), "transformation is not a 3D rigid-body transformation"

    scaled_rotation: np.ndarray = transformation[:3, :3]
    translation: np.ndarray = transformation[:3, 3]

    scale: float = np.linalg.norm(scaled_rotation, axis=1)[0]
    rotation: np.ndarray = scaled_rotation / scale

    return scale, rotation, translation


def rotation_matrix_to_euler(
    matrix: np.ndarray,
    degrees: bool = False,
    order: str = "ZYX",
) -> tuple[float, float, float]:
    """Converts a rotation matrix to a set of Euler angles."""

    assert matrix.shape == (
        3,
        3,
    ), "rotation is not a 3D rotation matrix"
    return Rotation.from_matrix(matrix).as_euler(order, degrees=degrees)


def rotation_matrix_to_vector(matrix: np.ndarray, degrees: bool = False) -> np.ndarray:
    """Converts a rotation matrix to a rotation vector."""

    assert matrix.shape == (
        3,
        3,
    ), "rotation is not a 3D rotation matrix"
    return Rotation.from_matrix(matrix).as_rotvec(degrees=degrees)



"""Module for functionality related to Hitnet disparity estimation model."""

from dataclasses import dataclass
from pathlib import Path
from typing import NamedTuple

import cv2
import numpy as np
import onnxruntime as onnxrt
import torch

from mynd.image import Image, PixelFormat, flip_image
from mynd.utils.containers import Pair

from .stereo_matcher import StereoMatcher


class SessionArgument(NamedTuple):
    """Class representing a session argument."""

    name: str
    shape: tuple
    type: type


@dataclass
class HitnetModel:
    """Class representing a Hitnet model."""

    session: onnxrt.InferenceSession

    @property
    def inputs(self) -> list[SessionArgument]:
        """Returns the inputs of the session."""
        arguments = self.session.get_inputs()
        return [
            SessionArgument(argument.name, tuple(argument.shape), argument.type)
            for argument in arguments
        ]

    @property
    def outputs(self) -> list[SessionArgument]:
        """Returns the inputs of the session."""
        arguments = self.session.get_outputs()
        return [
            SessionArgument(argument.name, tuple(argument.shape), argument.type)
            for argument in arguments
        ]

    @property
    def input_size(self) -> tuple[int, int]:
        """Returns the expected input size for the model as (H, W)."""
        tensor_argument: SessionArgument = self.inputs[0]
        batch, channels, height, width = tensor_argument.shape
        return (height, width)


def load_hitnet(path: Path) -> HitnetModel:
    """Loads a Hitnet model from an ONNX file."""

    if not path.exists():
        raise ValueError(f"model path does not exist: {path}")
    if not path.suffix == ".onnx":
        raise ValueError(f"model path is not an ONNX file: {path}")

    providers: list = ["CUDAExecutionProvider", "CPUExecutionProvider"]
    sess_options: onnxrt.SessionOptions = onnxrt.SessionOptions()

    session: onnxrt.InferenceSession = onnxrt.InferenceSession(
        str(path),
        sess_options=sess_options,
        providers=providers,
    )

    # TODO: Add validation based on session input and output

    return HitnetModel(session=session)


def create_hitnet_matcher(path: Path) -> StereoMatcher:
    """Creates a Hitnet stereo matcher."""

    model: HitnetModel | str = load_hitnet(path)

    def match_stereo_hitnet(left: Image, right: Image) -> Pair[Image]:
        """Matches a pair of rectified stereo images with the Hitnet model."""
        return _compute_disparity(model, left, right)

    return match_stereo_hitnet


def _preprocess_images(
    model: HitnetModel, left: Image, right: Image
) -> tuple[np.ndarray, np.ndarray]:
    """Preprocess input images for HITNET."""

    assert left.dtype in [
        np.uint8,
        np.uint16,
        np.uint32,
    ], "invalid dtype for left image"
    assert right.dtype in [
        np.uint8,
        np.uint16,
        np.uint32,
    ], "invalid dtype for right image"

    # Convert left image to grayscale
    match left.pixel_format:
        case PixelFormat.RGB:
            left_array: np.ndarray = cv2.cvtColor(left.to_array(), cv2.COLOR_RGB2GRAY)
        case PixelFormat.BGR:
            left_array: np.ndarray = cv2.cvtColor(left.to_array(), cv2.COLOR_BGR2GRAY)
        case PixelFormat.GRAY:
            left_array: np.ndarray = left.to_array()
        case _:
            raise NotImplementedError(f"invalid image format: {left.pixel_format}")

    # Convert right image to grayscale
    match right.pixel_format:
        case PixelFormat.RGB:
            right_array: np.ndarray = cv2.cvtColor(right.to_array(), cv2.COLOR_RGB2GRAY)
        case PixelFormat.BGR:
            right_array: np.ndarray = cv2.cvtColor(right.to_array(), cv2.COLOR_BGR2GRAY)
        case PixelFormat.GRAY:
            right_array: np.ndarray = right.to_array()
        case _:
            raise NotImplementedError(f"invalid image format: {right.pixel_format}")

    assert len(model.inputs) == 1, f"invalid number of inputs: {len(model.inputs)}"
    assert len(model.outputs) == 1, f"invalid number of outputs: {len(model.outputs)}"

    height: int
    width: int
    height, width = model.input_size

    left_array: np.ndarray = cv2.resize(left_array, (width, height), cv2.INTER_AREA)
    right_array: np.ndarray = cv2.resize(right_array, (width, height), cv2.INTER_AREA)

    # Grayscale needs expansion to reach H,W,C.
    # Need to do that now because resize would change the shape.
    if left_array.ndim == 2:
        left_array: np.ndarray = np.expand_dims(left_array, axis=-1)
    if right_array.ndim == 2:
        right_array: np.ndarray = np.expand_dims(right_array, axis=-1)

    left_normalized: np.ndarray = left_array / np.iinfo(left_array.dtype).max
    right_normalized: np.ndarray = right_array / np.iinfo(right_array.dtype).max

    # -> H,W,C=2 or 6 , normalized to [0,1]
    tensor: np.ndarray = np.concatenate((left_normalized, right_normalized), axis=-1)
    # -> C,H,W
    tensor: np.ndarray = tensor.transpose(2, 0, 1)
    # -> B=1,C,H,W
    tensor = np.expand_dims(tensor, 0).astype(np.float32)

    return tensor


def _postprocess_disparity(
    disparity: np.ndarray, image: Image, flip: bool = False
) -> np.ndarray:
    """Postprocess the disparity map by resizing it to match the original image,
    adjusting the disparity with the width ratio, and optionally flipping the disparity
    horizontally."""

    # Squeeze disparity to a 2D array
    disparity: np.ndarray = np.squeeze(disparity)

    # Scale disparities by the width ratios between the original images and the disparity maps
    scale: float = float(image.width) / float(disparity.shape[1])
    disparity *= scale

    # Resize disparity maps to the original image sizes
    disparity: np.ndarray = cv2.resize(
        disparity, (image.width, image.height), cv2.INTER_AREA
    )

    # If enabled, flip disparity map around y-axis (horizontally)
    if flip:
        disparity: np.ndarray = cv2.flip(disparity, 1)

    return disparity


def _compute_disparity(model: HitnetModel, left: Image, right: Image) -> Pair[Image]:
    """Computes the disparity for a pair of stereo images. The images needs to be
    rectified prior to disparity estimation. Returns the left and right disparity as
    arrays with float32 values."""

    # Create tensor from flipped images to get left disparity
    flipped_left: Image = flip_image(left)
    flipped_right: Image = flip_image(right)

    tensor: np.ndarray = _preprocess_images(model, left, right)
    flipped_tensor: np.ndarray = _preprocess_images(model, flipped_right, flipped_left)

    left_outputs: list[np.ndarray] = model.session.run(
        ["reference_output_disparity"], {"input": tensor}
    )
    right_outputs: list[np.ndarray] = model.session.run(
        ["reference_output_disparity"], {"input": flipped_tensor}
    )

    # Since we estimate the right disparity from the flipped images, we need to flip the
    # right disparity map back to the same perspective as the original rigth image
    left_disparity: np.ndarray = _postprocess_disparity(
        left_outputs[0], left, flip=False
    )
    right_disparity: np.ndarray = _postprocess_disparity(
        right_outputs[0], right, flip=True
    )

    left_disparity_image: Image = Image.from_array(left_disparity, PixelFormat.X)
    right_disparity_image: Image = Image.from_array(right_disparity, PixelFormat.X)

    return Pair(first=left_disparity_image, second=right_disparity_image)



"""Module for incremental point cloud registrators."""

import numpy as np
import open3d

from mynd.geometry.point_cloud import PointCloud

import open3d.pipelines.registration as reg

from .types import RegistrationResult, PointCloudRefiner


"""
Worker functions:
 - register_regular_icp
 - register_colored_icp
"""


def register_regular_icp(
    source: PointCloud,
    target: PointCloud,
    transformation: np.ndarray,
    *,
    distance_threshold: float,
    estimation_method: reg.TransformationEstimation = reg.TransformationEstimationPointToPlane(),
    convergence_criteria: reg.ICPConvergenceCriteria = reg.ICPConvergenceCriteria(),
) -> RegistrationResult:
    """Registers the source to the target with ICP."""

    result: reg.RegistrationResult = reg.registration_icp(
        source=source,
        target=target,
        init=transformation,
        max_correspondence_distance=distance_threshold,
        estimation_method=estimation_method,
        criteria=convergence_criteria,
    )

    information: np.ndarray = reg.get_information_matrix_from_point_clouds(
        source=source,
        target=target,
        max_correspondence_distance=distance_threshold,
        transformation=result.transformation,
    )

    return RegistrationResult(
        fitness=result.fitness,
        inlier_rmse=result.inlier_rmse,
        correspondence_set=result.correspondence_set,
        transformation=result.transformation,
        information=information,
    )


def register_colored_icp(
    source: PointCloud,
    target: PointCloud,
    transformation: np.ndarray,
    *,
    distance_threshold: float,
    estimation_method: reg.TransformationEstimation = reg.TransformationEstimationForColoredICP(),
    convergence_criteria: reg.ICPConvergenceCriteria = reg.ICPConvergenceCriteria(),
) -> RegistrationResult:
    """Registers the source to the target with ICP."""

    result: reg.RegistrationResult = reg.registration_colored_icp(
        source=source,
        target=target,
        init=transformation,
        max_correspondence_distance=distance_threshold,
        estimation_method=estimation_method,
        criteria=convergence_criteria,
    )

    information: np.ndarray = reg.get_information_matrix_from_point_clouds(
        source=source,
        target=target,
        max_correspondence_distance=distance_threshold,
        transformation=result.transformation,
    )

    return RegistrationResult(
        fitness=result.fitness,
        inlier_rmse=result.inlier_rmse,
        correspondence_set=result.correspondence_set,
        transformation=result.transformation,
        information=information,
    )


"""
Factory functions:
 - create_convergence_criteria_icp
 - create_estimator_point_to_plane
 - create_estimator_icp_generalized
 - create_estimator_icp_colored
 - create_icp_registrator_regular
 - create_icp_registrator_colored
 - create_kernel_loss_huber
 - create_kernel_loss_tukey
"""


def create_convergence_criteria_icp(
    max_iteration: int = 30,
    relative_fitness: float = 1e-06,
    relative_rmse: float = 1e-06,
) -> reg.ICPConvergenceCriteria:
    """Creates a ICP convergence criteria."""
    return reg.ICPConvergenceCriteria(
        max_iteration=max_iteration,
        relative_fitness=relative_fitness,
        relative_rmse=relative_rmse,
    )


def create_estimator_point_to_plane(
    kernel: reg.RobustKernel | None = None,
) -> reg.TransformationEstimation:
    """Creates a point to plane transformation estimator."""
    return reg.TransformationEstimationPointToPlane(kernel=kernel)


def create_estimator_icp_generalized(
    epsilon: float,
    kernel: reg.RobustKernel | None = None,
) -> reg.TransformationEstimation:
    """Creates a generalized ICP transformation estimator."""
    return reg.TransformationEstimationForGeneralizedICP(epsilon=epsilon, kernel=kernel)


def create_estimator_icp_colored(
    lambda_geometric: float,
    kernel: reg.RobustKernel | None = None,
) -> reg.TransformationEstimation:
    """Creates a colored ICP transformation estimator."""
    return reg.TransformationEstimationForColoredICP(
        lambda_geometric=lambda_geometric,
        kernel=kernel,
    )


def create_icp_registrator_regular(
    estimation_method: reg.TransformationEstimation,
    convergence_criteria: reg.ICPConvergenceCriteria,
    distance_threshold: float,
) -> PointCloudRefiner:
    """Creates a regular ICP registrator from the given arguments."""

    def regular_icp_wrapper(
        source: PointCloud,
        target: PointCloud,
        transformation: np.ndarray,
    ) -> RegistrationResult:
        """Closure wrapper for regular ICP registration method."""
        return register_regular_icp(
            source=source,
            target=target,
            transformation=transformation,
            estimation_method=estimation_method,
            convergence_criteria=convergence_criteria,
            distance_threshold=distance_threshold,
        )

    return regular_icp_wrapper


def create_icp_registrator_colored(
    estimation_method: reg.TransformationEstimation,
    convergence_criteria: reg.ICPConvergenceCriteria,
    distance_threshold: float,
) -> PointCloudRefiner:
    """Creates a colored ICP registrator from the given arguments."""

    def colored_icp_wrapper(
        source: PointCloud,
        target: PointCloud,
        transformation: np.ndarray,
    ) -> RegistrationResult:
        """Closure wrapper for colored ICP registration method."""
        return register_colored_icp(
            source=source,
            target=target,
            transformation=transformation,
            estimation_method=estimation_method,
            convergence_criteria=convergence_criteria,
            distance_threshold=distance_threshold,
        )

    return colored_icp_wrapper


def create_kernel_loss_huber(k: float) -> reg.RobustKernel:
    """Creates a robust kernel with Huber loss."""
    return reg.HuberLoss(k=k)


def create_kernel_loss_tukey(k: float) -> reg.RobustKernel:
    """Creates a robust kernel with Tukey loss."""
    return reg.TukeyLoss(k=k)



"""Module for environment file and values."""

import os
import pathlib
import typing

import dotenv


ROOT_DIR: pathlib.Path = pathlib.Path(
    os.path.dirname(os.path.abspath(__file__))
).parents[2]
ENV_PATH: pathlib.Path = ROOT_DIR / pathlib.Path(".env")


def values() -> dict[str, typing.Any]:
    """Returns the keys and values in the environment file."""
    return dotenv.dotenv_values(ENV_PATH)


def get_value(key: str) -> typing.Any | None:
    """Returns the environment value for the given key."""
    return dotenv.dotenv_values(ENV_PATH).get(key)


def get_path() -> pathlib.Path:
    """Returns the environment file path."""
    return ENV_PATH



"""Unit tests for the mynds custom containers."""

import pytest

from mynd.utils.containers import Pair, Registry


# Tests for Pair class
def test_pair_creation():
    pair = Pair(1, 2)
    assert pair.first == 1
    assert pair.second == 2


def test_pair_immutability():
    pair = Pair("a", "b")
    with pytest.raises(AttributeError):
        pair.first = "c"


def test_pair_with_different_types():
    pair = Pair[object](1, "a")
    assert isinstance(pair.first, int)
    assert isinstance(pair.second, str)


# Tests for Registry class
@pytest.fixture
def sample_registry():
    registry = Registry[str, int]()
    registry.insert("one", 1)
    registry.insert("two", 2)
    registry.insert("three", 3)
    return registry


def test_registry_len(sample_registry):
    assert len(sample_registry) == 3
    assert sample_registry.count == 3


def test_registry_contains(sample_registry):
    assert "one" in sample_registry
    assert "four" not in sample_registry


def test_registry_getitem(sample_registry):
    assert sample_registry["two"] == 2
    with pytest.raises(KeyError):
        _ = sample_registry["four"]


def test_registry_setitem():
    registry = Registry[str, str]()
    registry["key"] = "value"
    assert registry["key"] == "value"


def test_registry_values(sample_registry):
    assert set(sample_registry.values) == {1, 2, 3}


def test_registry_keys(sample_registry):
    assert set(sample_registry.keys) == {"one", "two", "three"}


def test_registry_get(sample_registry):
    assert sample_registry.get("two") == 2
    assert sample_registry.get("four") is None
    assert sample_registry.get("four", 0) == 0


def test_registry_items(sample_registry):
    items = list(sample_registry.items())
    assert ("one", 1) in items
    assert ("two", 2) in items
    assert ("three", 3) in items


def test_registry_insert():
    registry = Registry[int, str]()
    registry.insert(1, "one")
    assert registry[1] == "one"


def test_registry_remove(sample_registry):
    sample_registry.remove("two")
    assert "two" not in sample_registry
    assert len(sample_registry) == 2


def test_registry_pop(sample_registry):
    value = sample_registry.pop("two")
    assert value == 2
    assert "two" not in sample_registry
    assert len(sample_registry) == 2


def test_registry_pop_nonexistent():
    registry = Registry[str, int]()
    with pytest.raises(KeyError):
        registry.pop("nonexistent")



"""Module for building registration pipelines."""

from collections.abc import Callable
from typing import Any, TypeAlias

import open3d.pipelines.registration as reg

from mynd.geometry.point_cloud import PointCloud, PointCloudProcessor
from mynd.geometry.point_cloud import create_downsampler, create_normal_estimator

from .feature_registrators import (
    create_fpfh_extractor,
    create_point_to_point_estimator,
    create_correspondence_validators,
    create_convergence_criteria_ransac,
    create_ransac_registrator,
)

from .icp_registrators import (
    create_kernel_loss_huber,
    create_kernel_loss_tukey,
    create_estimator_point_to_plane,
    create_estimator_icp_colored,
    create_convergence_criteria_icp,
    create_icp_registrator_regular,
    create_icp_registrator_colored,
)

from .types import (
    FeatureExtractor,
    PointCloudAligner,
    PointCloudRefiner,
)

from .pipeline import RegistrationPipeline


Pipeline: TypeAlias = RegistrationPipeline


def build_registration_pipeline(config: dict) -> Pipeline:
    """Builds a registration pipeline from the given config."""

    ALIGNER_KEY: str = "aligner"
    REFINER_KEY: str = "refiner"

    assert ALIGNER_KEY in config, f"missing required key: {ALIGNER_KEY}"
    assert REFINER_KEY in config, f"missing required key: {REFINER_KEY}"

    aligner_module: Pipeline.AlignerModule = _build_aligner_module(
        config.get(ALIGNER_KEY)
    )

    refiner_modules: list[Pipeline.RefinerModule] = [
        _build_refiner_module(section) for section in config.get(REFINER_KEY)
    ]

    return Pipeline(aligner_module, refiner_modules)


def _build_aligner_module(config: dict) -> Pipeline.AlignerModule:
    """Builds an aligner module from the configuration."""

    MATCHER_FACTORIES: list[str] = {
        "feature_ransac": build_ransac_registrator,
    }

    matcher_type: str = config.get("type")
    matcher_params: dict = config.get("matcher")

    preprocessor: PointCloudProcessor = build_point_cloud_processor(
        config.get("preprocessor")
    )

    if matcher_type not in MATCHER_FACTORIES:
        raise NotImplementedError(
            f"invalid alignment matcher - valid options are: {MATCHER_FACTORIES.keys()}"
        )

    factory = MATCHER_FACTORIES.get(matcher_type)
    matcher: PointCloudAligner = factory(matcher_params)

    return Pipeline.AlignerModule(preprocessor, matcher)


def _build_refiner_module(config: dict) -> Pipeline.RefinerModule:
    """Builds an refiner module from the configuration."""

    REFINER_FACTORIES: dict[str, Callable] = {
        "regular_icp": build_regular_icp_registrator,
        "colored_icp": build_colored_icp_registrator,
    }

    preprocessor: PointCloudProcessor = build_point_cloud_processor(
        config.get("preprocessor")
    )

    matcher_type: str = config.get("type")
    matcher_params: str = config.get("matcher")

    factory = REFINER_FACTORIES.get(matcher_type)
    matcher: PointCloudRefiner = factory(matcher_params)

    return Pipeline.RefinerModule(preprocessor, matcher)


def build_point_cloud_processor(components: dict[str, Any]) -> PointCloudProcessor:
    """Builds a point cloud preprocessor from a configuration."""

    processors: list[PointCloudProcessor] = list()

    if "downsample" in components:
        processors.append(create_downsampler(**components.get("downsample")))

    if "estimate_normals" in components:
        processors.append(create_normal_estimator(**components.get("estimate_normals")))

    def preprocess_point_cloud(cloud: PointCloud) -> PointCloud:
        """Preprocesses a point cloud."""
        processed: PointCloud = cloud
        for processor in processors:
            processed: PointCloud = processor(processed)

        return processed

    return preprocess_point_cloud


def build_ransac_registrator(components: dict[str, Any]) -> PointCloudAligner:
    """Builds a RANSAC registrator from a collection of parameters."""

    for key in [
        "feature",
        "point_to_point",
        "validators",
        "convergence",
        "algorithm",
    ]:
        assert key in components, f"missing build component: {key}"

    feature_extractor: FeatureExtractor = create_fpfh_extractor(
        **components.get("feature")
    )
    estimation_method = create_point_to_point_estimator(
        **components.get("point_to_point")
    )
    validators = create_correspondence_validators(**components.get("validators"))
    convergence_criteria = create_convergence_criteria_ransac(
        **components.get("convergence")
    )

    return create_ransac_registrator(
        feature_extractor=feature_extractor,
        estimation_method=estimation_method,
        validators=validators,
        convergence_criteria=convergence_criteria,
        **components.get("algorithm"),
    )


def build_regular_icp_registrator(
    parameters: dict,
) -> PointCloudRefiner:
    """Builds a regular ICP registrator from a configuration."""

    if BUILD_HUBER_KEY in parameters:
        kernel: reg.RobustKernel = create_kernel_loss_huber(
            **parameters.get(BUILD_HUBER_KEY)
        )
    elif BUILD_TUKEY_KEY in parameters:
        kernel: reg.RobustKernel = create_kernel_loss_tukey(
            **parameters.get(BUILD_TUKEY_KEY)
        )
    else:
        kernel = None

    # NOTE: For now we only use a point to plane estimator for regular ICP
    estimator: reg.TransformationEstimation = create_estimator_point_to_plane(
        kernel=kernel
    )

    if CONVERGENCE_CRITERIA_KEY in parameters:
        convergence_criteria: reg.ICPConvergenceCriteria = (
            create_convergence_criteria_icp(**parameters.get(CONVERGENCE_CRITERIA_KEY))
        )
    else:
        convergence_criteria: reg.ICPConvergenceCriteria = (
            create_convergence_criteria_icp()
        )

    if DISTANCE_THRESHOLD_KEY not in parameters:
        raise ValueError(f"regular icp builder: missing key '{DISTANCE_THRESHOLD_KEY}'")

    distance_threshold: float = parameters.get("distance_threshold")

    return create_icp_registrator_regular(
        estimation_method=estimator,
        convergence_criteria=convergence_criteria,
        distance_threshold=distance_threshold,
    )


# TODO: Add named tuple to hold build keys
BUILD_HUBER_KEY: str = "huber_kernel"
BUILD_TUKEY_KEY: str = "tukey_kernel"
CONVERGENCE_CRITERIA_KEY: str = "convergence_criteria"
DISTANCE_THRESHOLD_KEY: str = "distance_threshold"


def build_colored_icp_registrator(
    parameters: dict,
) -> PointCloudRefiner:
    """Builds an incremental registrator from a configuration."""

    COLOR_ESTIMATOR_KEY: str = "colored_icp_estimation"

    if BUILD_HUBER_KEY in parameters:
        kernel: reg.RobustKernel = create_kernel_loss_huber(
            **parameters.get(BUILD_HUBER_KEY)
        )
    elif BUILD_TUKEY_KEY in parameters:
        kernel: reg.RobustKernel = create_kernel_loss_tukey(
            **parameters.get(BUILD_TUKEY_KEY)
        )
    else:
        kernel = None

    if COLOR_ESTIMATOR_KEY in parameters:
        estimator: reg.TransformationEstimation = create_estimator_icp_colored(
            **parameters.get(COLOR_ESTIMATOR_KEY),
            kernel=kernel,
        )
    else:
        raise ValueError(f"colored icp builder: missing key '{COLOR_ESTIMATOR_KEY}'")

    if CONVERGENCE_CRITERIA_KEY in parameters:
        criteria: reg.ICPConvergenceCriteria = create_convergence_criteria_icp(
            **parameters.get(CONVERGENCE_CRITERIA_KEY),
        )

    if DISTANCE_THRESHOLD_KEY not in parameters:
        raise ValueError(f"colored icp builder: missing key '{DISTANCE_THRESHOLD_KEY}'")

    distance_threshold: float = parameters.get(DISTANCE_THRESHOLD_KEY)

    return create_icp_registrator_colored(
        estimation_method=estimator,
        convergence_criteria=criteria,
        distance_threshold=distance_threshold,
    )



"""Module for point cloud IO, i.e. reading and writing point clouds."""

from pathlib import Path

import open3d

from .point_cloud_types import PointCloud, PointCloudLoader


def read_point_cloud(path: str | Path) -> PointCloud | str:
    """Loads a point cloud from a file."""
    try:
        point_cloud: PointCloud = open3d.io.read_point_cloud(str(path))
        return point_cloud
    except IOError as error:
        return str(error)


def create_point_cloud_loader(source: str | Path) -> PointCloudLoader:
    """Creates a point cloud loader for the source."""

    def load_point_cloud() -> PointCloud | str:
        """Loads a point cloud from the source."""
        return read_point_cloud(path=source)

    return load_point_cloud
